import subprocess
import networkx as nx
import tempfile
import inspect
import os
import logging
import subprocess
import time
import importlib
import numpy as np
import psutil

from aspmc.graph.hypergraph import Hypergraph

from aspmc.util import *

import aspmc.compile.constrained_compile as concom
from aspmc.compile.constrained_sdd import ConstrainedSDD
from aspmc.compile.constrained_ddnnf import ConstrainedDDNNF
from aspmc.compile.circuit import Circuit
import aspmc.compile.dtree as dtree
import aspmc.compile.vtree as vtree

import aspmc.signal_handling as my_signals

import aspmc.config as config

src_path = os.path.abspath(os.path.realpath(inspect.getfile(inspect.currentframe())))
src_path = os.path.realpath(os.path.join(src_path, '../../external'))

logger = logging.getLogger("aspmc")

class SolvingError(Exception):
    '''raise this when a solver errors'''


class CNF(object):
    """This class is an extended cnf class, which can be used to compile and 
    evaluate algebraic model counting problems over cnfs using the knowledge compiler in aspmc.config.

    The syntax of extended cnf files is as follows:

        * format string: one line specifying the number of variables and clauses.

            `p cnf <nr_vars> <len(clauses)>`
        * clauses: one line for each clause `c = [lit_1, lit_2, ..., lit_n]`.

            `<lit_1> <lit_2> ... <lit_n> 0`
        * auxilliary: one line specifying a list `[v_1, ..., v_n]` of variables that may be projected away without changing the evaluation result.
        
            `c p auxilliary <v_1> <v_2> ... <v_n> 0`
        * weights: one line for each weight "w = np.array([w_1, w_2, ..., w_n])" of a literal `lit`.

            `c p weight <lit> <w_1>;<w_2>;...;<w_n> 0`
        * semirings: one line containing all the names of the semiring modules `semirings = [s_1, s_2, ..., s_n]`.

            `c p semirings <s_1.__name__> <s_2.__name__> ... <s_n.__name__> 0`
        * quantified: one line for each list `q_i = [v_1, ..., v_n]` in quantified. Lowest index quantified variables first.

            `c p quantify <v_1> <v_2> ... <v_n> 0`
        * tranform: one line containing the transformation function from semiring[1] to semiring[0] in string representation.
            Must be such that `eval(repr(transform))` works independently of the imported modules.
            
            `c p transform <repr(transform)> 0`

    Args:
        path (:obj:`string`, optional): Optional parameter specifying the location of an extended cnf file to load. Defaults to None.

    Attributes:
        clauses (list): A list of clauses with literals in minisat format.
        nr_vars (int): The number of variables that the cnf is specified over.
        auxilliary (set): A set of variables that may be projected away without changing the evaluation result.
        weights (dict): A dictionary that can contain for each integer from {-nr_vars, ..., -1, 1, ..., nr_vars} a weight over a semiring.
            Note that even if there is only one semiring value it must be encapsulated in a numpy array.
        semirings (list): A list of semiring modules that are used in this cnf.
            See aspmc.semirings for how these modules should look.
            Currently at most two semirings are supported.
        quantified (list): A list of lists of variables (integers) specifying, which variables are "quantified" over which semirings.
            Must have the same length as `semirings`.
            The variables in quantified[i] are over semirings[i].
            The variables in quantified[i] are quantified before the ones in quantified[i+i].
        transform (string): A string representation of a python function that takes a values from semiring[i] and returns a value `x` that can be given to 
            `semirings[i-1].from_value(x)` to obtain a value in semirings[i-1].

    """
    def __init__(self, path = None, string = None):
        assert(path is None or string is None)
        self.clauses = []
        self.nr_vars = 0
        self.auxilliary = set()
        self.weights = {}
        self.semirings = []
        self.quantified = []
        self.transforms = []
        if path is not None:
            with open(path) as in_file:
                for line in in_file:
                    line = line.split()
                    if len(line) == 0:
                        continue
                    if line[0] == 'c':
                        if len(line) > 2 and line[1] == 'p':
                            if line[2] == "weight":
                                if int(line[3]) != 0:
                                    self.weights[int(line[3])] = ' '.join(line[4:-1])
                            elif line[2] == "semirings":
                                self.semirings = [ importlib.import_module(mod) for mod in line[3:-1] ]
                            elif line[2] == "transform":
                                self.transforms.append(' '.join(line[3:-1]))
                            elif line[2] == "quantify":
                                self.quantified.append([int(x) for x in line[3:-1]])
                            elif line[2] == "auxilliary":
                                self.auxilliary.update([int(x) for x in line[3:-1]])
                            else:
                                logger.error(f"Unknown property {line[2]}!")
                            if line[-1] != '0':
                                logger.error("Property line not ended with 0!")
                    elif line[0] == 'p':
                        self.nr_vars = int(line[2])
                    else:
                        line = [int(l) for l in line]
                        self.clauses.append(line[:-1])
                        
        if string is not None:
            for line in string.split("\n"):
                line = line.split()
                if len(line) == 0:
                    continue
                if line[0] == 'c':
                    if len(line) > 2 and line[1] == 'p':
                        if line[2] == "weight":
                            self.weights[int(line[3])] = ' '.join(line[4:-1])
                        elif line[2] == "semirings":
                            self.semirings = [ importlib.import_module(mod) for mod in line[3:-1] ]
                        elif line[2] == "transform":
                            self.transforms.append(' '.join(line[3:-1]))
                        elif line[2] == "quantify":
                            self.quantified.append([int(x) for x in line[3:-1]])
                        elif line[2] == "auxilliary":
                                self.auxilliary.update([int(x) for x in line[3:-1]])
                        else:
                            logger.error(f"Unknown property {line[2]}!")
                        if line[-1] != '0':
                            logger.error("Property line not ended with 0!")
                elif line[0] == 'p':
                    self.nr_vars = int(line[2])
                else:
                    line = [int(l) for l in line]
                    self.clauses.append(line[:-1])

        # check whether the input is reasonable
        if len(self.quantified) != len(self.semirings):
            logger.error("We must have the same number of semirings and quantifiers!")
            exit(-1)
        if len(self.semirings) != len(self.transforms) + 1 and (path is not None or string is not None):
            logger.error("If there are multiple semirings, we need a transforms between them.")
            exit(-1)
        if len(self.weights) > 0 and len(self.quantified) == 0:
            self.quantified = [ set(range(1,self.nr_vars + 1)) ]
            import aspmc.semirings.probabilistic
            self.semirings = [ aspmc.semirings.probabilistic ]
            
        var_to_val_type = [ len(self.semirings) - 1 for _ in range(self.nr_vars + 1) ]
        for val_type, vars in enumerate(self.quantified):
            for var in vars:
                var_to_val_type[var] = val_type
                
        for idx in self.weights:
            val_type = var_to_val_type[idx]
            self.weights[idx] = np.array([ self.semirings[val_type].parse(w) for w in self.weights[idx].split(";") ])

    def __repr__(self):
        return str(self)

    def __str__(self):
        ret = f"p cnf {self.nr_vars} {len(self.clauses)}\n"
        for c in self.clauses:
            ret += f"{' '.join([str(l) for l in c])} 0\n"
        for idx in self.weights:
            if abs(idx) in self.quantified[0]:
                weight = ';'.join([ self.semirings[0].to_string(w) for w in self.weights[idx] ])
            else:
                weight = ';'.join([ self.semirings[1].to_string(w) for w in self.weights[idx] ])
            ret += f"c p weight {idx} {weight} 0\n"
        if len(self.semirings) > 0:
            ret += f"c p semirings {' '.join([ x.__name__ for x in self.semirings])} 0\n"
        for transform in self.transforms:
            ret += f"c p transform {transform} 0\n"
        for l in self.quantified:
            ret += f"c p quantify {' '.join([str(x) for x in l])} 0\n"
        ret += f"c p auxilliary {' '.join([str(x) for x in self.auxilliary])} 0\n"
        return ret

    def write_kc_cnf(self, out_file):
        out_file.write(f"p cnf {self.nr_vars} {len(self.clauses)}\n".encode())
        for c in self.clauses:
            out_file.write(f"{' '.join([str(l) for l in c])} 0\n".encode())
        for idx in range(1, self.nr_vars + 1):
            if idx not in self.auxilliary:
                out_file.write(f"c p weight {idx} {idx} 0\n".encode())
                out_file.write(f"c p weight {-idx} {-idx} 0\n".encode())

    def write_maxsat_cnf(self, out_file):
        import math
        real_weights = { l : w[0].value for l,w in self.weights.items() }
        if self.semirings[0].__name__ == "aspmc.semirings.maxtimes":
            # we first need to convert into maxplus weights
            real_weights = { l : math.log(w) if w > 0 else float("-inf") for l,w in real_weights.items() }
        elif self.semirings[0].__name__ == "aspmc.semirings.minplus":
            # we first need to convert into maxplus weights
            real_weights = { l : -w if w != float("inf") else float("-inf") for l,w in real_weights.items() }
        elif self.semirings[0].__name__ != "aspmc.semirings.maxplus":
            logger.error(f"MaxSAT evaluation is currently not supported for semiring {self.semirings[0].__name__}")
            exit(-1)
        # sort out variables that are irrelevant
        real_weights = { l : w for l,w in real_weights.items() if w != real_weights[-l]}
        # handle hard constraints from literals with weight -inf and keep the rest
        negated_units = [ l for l,w in real_weights.items() if w == float('-inf')]
        real_weights = { l : w for l,w in real_weights.items() if w != float('-inf')}
        # make sure every variable has exactly one weight and that weight is greater than 0
        for i in range(1, self.nr_vars + 1):
            if i in real_weights:
                if -i in real_weights:
                    if real_weights[i] < real_weights[-i]:
                        real_weights[-i] -= real_weights[i]
                        del real_weights[i]
                    else:
                        real_weights[i] -= real_weights[-i]
                        del real_weights[-i]
                else:
                    if real_weights[i] < 0:
                        real_weights[-i] = -real_weights[i]
                        del real_weights[i]
            else:
                if -i in real_weights:
                    if real_weights[-i] < 0:
                        real_weights[i] = -real_weights[-i]
                        del real_weights[-i]
        max_exp = max([ math.ceil(-math.log10(w)) for w in real_weights.values() if w > 0 ] + [0])
        real_weights = { l : math.floor(w*10**(8 + max_exp)) for l,w in real_weights.items() if abs(w) >= 0.1**(8 + max_exp) }
        gcd = 0
        for w in real_weights.values():
            gcd = math.gcd(w, gcd)
        real_weights = { l : w//gcd for l,w in real_weights.items() }
        top = sum(real_weights.values()) + 2
        if top >= 2**63:
            logger.error(f"Cannot reduce this instance to a maxsat instance.")
            exit(-1)
        out_file.write(f"p wcnf {self.nr_vars} {len(self.clauses) + len(real_weights)} {top}\n".encode())
        for c in self.clauses:
            out_file.write(f"{top} {' '.join([str(l) for l in c])} 0\n".encode())
        for l in negated_units:
            out_file.write(f"{top} {-l} 0\n".encode())
        for l, w in real_weights.items():
            out_file.write(f"{w} {l} 0\n".encode())

        #c = CNF()
        #c.clauses = self.clauses
        #import aspmc.semirings.maxplus as maxplus
        #c.weights = { i : np.array([maxplus.one()]) for i in range(-self.nr_vars, self.nr_vars + 1)}
        #del c.weights[0]
        #c.semirings = [maxplus]
        #c.quantified = [list(range(1,self.nr_vars + 1))]
        #for l, w in real_weights.items():
        #    c.weights[l] = np.array([maxplus.MaxPlusFloat(-w)])
        #c.nr_vars = self.nr_vars
        #print(c)


    def get_defined(self, P, timeout = "150"):
        """Figures out the subset of variables of the cnf that are defined by `P` w.r.t. the cnf.
        Calls a C++ binary for performance reasons.

        Args:
            P (iterable): The set of input variables that can be used for definitions.

        Returns:
            list: The list of variables that are defined by the inputs `P` w.r.t. the cnf.
        """
        (cnf_fd, cnf_tmp) = tempfile.mkstemp()
        my_signals.tempfiles.add(cnf_tmp)
        (input_fd, input_tmp) = tempfile.mkstemp()
        my_signals.tempfiles.add(input_tmp)
        with os.fdopen(cnf_fd, 'wb') as cnf_file:
            self.to_stream(cnf_file)
        with os.fdopen(input_fd, 'w') as input_file:
            input_file.write(" ".join([str(p) for p in list(P) + [0]]))
        p = subprocess.Popen(["timeout",  timeout, os.path.join(src_path, "minisat-definitions/bin/defined"), cnf_tmp, input_tmp], stdout=subprocess.PIPE)
        p.wait()
        ret = p.stdout.read().decode().split(' ')[:-1]
        ret = [ int(v) for v in ret ]
        p.stdout.close()
        os.remove(cnf_tmp)
        my_signals.tempfiles.remove(cnf_tmp)
        os.remove(input_tmp)
        my_signals.tempfiles.remove(input_tmp)
        return ret
        
    def is_sat(self):
        """Calls `minisat` to check if the cnf is satisfiable.

        Returns:
            bool: `True` if the cnf is satisfiable, `False` otherwise.
        """
        p = subprocess.Popen([os.path.join(src_path, "minisat-definitions/bin/minisat")], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        p.stdin.write(str(self).encode())
        p.stdin.close()
        p.wait()
        for line in p.stdout.read().decode().split('\n'):
            line = line.split()
            if len(line) == 0:
                continue
            if line[0] == 'UNSATISFIABLE':
                p.stdout.close()
                return False
            elif line[0] == 'SATISFIABLE':
                p.stdout.close()
                return True

    def to_file(self, path, extras = False):
        """Write the cnf to the file with the name `path`.

        Args:
            path (string): The path of the file the cnf should be written to.
            extras (bool, optional): Whether the extra information like weights should be written. Defaults to False.

        Returns:
            None
        """
        with open(path, mode = 'w') as file_out:
            file_out.write(f"p cnf {self.nr_vars} {len(self.clauses)}\n")
            for c in self.clauses:
                file_out.write(f"{' '.join([str(l) for l in c])} 0\n")
            if extras:
                for idx in self.weights:
                    if abs(idx) in self.quantified[0]:
                        weight = ';'.join([ self.semirings[0].to_string(w) for w in self.weights[idx] ])
                    else:
                        weight = ';'.join([ self.semirings[1].to_string(w) for w in self.weights[idx] ])
                    file_out.write(f"c p weight {idx} {weight} 0\n")
                if len(self.semirings) > 0:
                    file_out.write(f"c p semirings {' '.join([ x.__name__ for x in self.semirings])} 0\n")
                for transform in self.transforms:
                    file_out.write(f"c p transform {transform} 0\n")
                for l in self.quantified:
                    file_out.write(f"c p quantify {' '.join([str(x) for x in l])} 0\n")
                file_out.write(f"c p auxilliary {' '.join([str(x) for x in self.auxilliary])} 0\n")
            

    def to_stream(self, stream, extras = False):
        """Write the cnf to the stream `stream`.

        Args:
            stream (stream): The stream the cnf should be written to. Must accept binary encoding.
            extras (bool, optional): Whether the extra information like weights should be written. Defaults to False.

        Returns:
            None
        """            
        stream.write(f"p cnf {self.nr_vars} {len(self.clauses)}\n".encode())
        for c in self.clauses:
            stream.write(f"{' '.join([str(l) for l in c])} 0\n".encode())
        if extras:
            for idx in self.weights:
                if abs(idx) in self.quantified[0]:
                    weight = ';'.join([ self.semirings[0].to_string(w) for w in self.weights[idx] ])
                else:
                    weight = ';'.join([ self.semirings[1].to_string(w) for w in self.weights[idx] ])
                stream.write(f"c p weight {idx} {weight} 0\n".encode())
            if len(self.semirings) > 0:
                stream.write(f"c p semirings {' '.join([ x.__name__ for x in self.semirings])} 0\n".encode())
            for transform in self.transforms:
                stream.write(f"c p transform {transform} 0\n".encode())
            for l in self.quantified:
                stream.write(f"c p quantify {' '.join([str(x) for x in l])} 0\n".encode())
            stream.write(f"c p auxilliary {' '.join([str(x) for x in self.auxilliary])} 0\n".encode())

    def primal_graph(self):
        """Construct the an `nx.Graph` that corresponds to the primal graph of the cnf.

        Returns:
            nx.Graph: The primal graph of the cnf. 
        """
        graph = nx.Graph()
        graph.add_nodes_from(range(1, self.nr_vars+1))
        for c in self.clauses:
            graph.add_edges_from(sum([[(abs(l),abs(lp)) for l in c if abs(l) != abs(lp)] for lp in c], []))
        return graph

    def primal_hypergraph(self):        
        """Construct the an `aspmc.graph.Hypergraph` that corresponds to the primal hypergraph of the cnf.

        Returns:
            aspmc.graph.Hypergraph: The primal hypergraph of the cnf. 
        """
        graph = Hypergraph()
        graph.add_nodes_from(range(1, self.nr_vars+1))
        for c in self.clauses:
            graph.add_edge([ abs(l) for l in c ])
        return graph

    def get_weights(self):
        """Get some relevant information about the weights of the cnf in a convenient format.

        Returns:
            (list, object, object, type): 
            
            The weights of the literals. `weights[2*(i-1)]` is the weight of `i` and `weights[2*(i-1) + 1]` is the weight of `-i`.
            
            The zero of the (outermost) AMC instance.
            
            The one of the (outermost) AMC instance.
            
            The type of the weights that should be used for numpy arrays.
        """
        if len(self.semirings) == 0:
            weights = [ np.array([ 1 ], dtype = object) for _ in range(self.nr_vars*2) ]
            zero = 0
            one = 1
            dtype = object
        elif len(self.semirings) >= 1:
            weights = []
            for i in range(len(self.weights)):
                weights.append(self.weights[to_dimacs(i)])
            zero = self.semirings[0].zero()
            one = self.semirings[0].one()
            dtype = self.semirings[0].dtype
        return (weights, zero, one, dtype)

    def remove_trivial_clauses(self):
        """Removes all the trivial clauses from the cnf. Trivial clauses are those clauses that contain both `v` and `-v` for some variables `v`.

        Returns:
            None
        """
        new_clauses = []
        for c in self.clauses:
            if not any([ any([ c[i] == -c[j] for j in range(i+1, len(c)) ]) for i in range(len(c)) ]):
                new_clauses.append(c)
        self.clauses = new_clauses


    def evaluate_trivial(self):
        """Checks if this is a trivial instance and if so returns its value. Before the check all the trivial clauses are removed.

        An instance is trivial if one of the following is true:
            * It contains no clauses.
            * It is unsatisfiable.

        Returns:
            object: The value of the AMC instance if it is trivial and `None` otherwise.
        """
        self.remove_trivial_clauses()
        if len(self.semirings) == 0:
            if len(self.clauses) == 0:
                if self.nr_vars == 0:
                    return [ 1 ]
                else:
                    weights, zero, one, dtype = self.get_weights()
                    first_shape = (np.shape(weights[0])[0], ) + np.shape(one)
                    res = np.empty(first_shape, dtype=dtype)
                    res[:] = one
                    for i in range(self.nr_vars):
                        res *= weights[to_pos(i)] + weights[neg(to_pos(i))]
                    return res
            elif not self.is_sat():
                return [ 0 ]
        elif len(self.semirings) == 1:
            if len(self.clauses) == 0:
                if self.nr_vars == 0:
                    return [ self.semirings[0].one() ]
                else:
                    weights, zero, one, dtype = self.get_weights()
                    first_shape = (np.shape(weights[0])[0], ) + np.shape(one)
                    res = np.empty(first_shape, dtype=dtype)
                    res[:] = one
                    for i in range(self.nr_vars):
                        res *= weights[to_pos(i)] + weights[neg(to_pos(i))]
                    return res
            elif not self.is_sat():        
                weights, zero, one, dtype = self.get_weights()
                first_shape = (np.shape(weights[0])[0], ) + np.shape(one)
                res = np.empty(first_shape, dtype=dtype)
                res[:] = zero
                return res
        elif len(self.semirings) == 2:
            if len(self.clauses) == 0:
                if self.nr_vars == 0:
                    return [ self.semirings[0].one() ]
                else:
                    weights, _, _, _ = self.get_weights()
                    second_shape = (np.shape(weights[0])[0], ) + np.shape(self.semirings[1].one())
                    res = np.empty(second_shape, dtype=self.semirings[1].dtype)
                    res[:] = self.semirings[1].one()
                    first = set(self.quantified[0])
                    second = set(range(1, self.nr_vars + 1))
                    second.difference_update(first)
                    for i in second:
                        res *= weights[to_pos(i)] + weights[neg(to_pos(i))]
                    f_transform = eval(self.transforms[0])
                    transform = lambda x : self.semirings[0].from_value(f_transform(x))
                    res = np.array([ transform(w) for w in res ], dtype = self.semirings[0].dtype)
                    for i in first:
                        res *= weights[to_pos(i)] + weights[neg(to_pos(i))]
                    return res
            elif not self.is_sat():
                first_shape = (np.shape(self.weights[0])[0], ) + np.shape(self.semirings[0].one())
                res = np.empty(first_shape, dtype=self.semirings[0].dtype)
                res[:] = self.semirings[0].zero()
                return res
        # TODO: implement for more semirings
                        
    
    @staticmethod
    def compile_single(file_name, knowledge_compiler = "c2d"):
        """Compiles a CNF into a tractable circuit. The output circuit is in the file `file_name + ".nnf"`.

        Currently supports c2d, miniC2D, d4 and sharpsat-td as knowledge compilers. 

        For c2d assumes that there is:
            * a cnf file `file_name`.
            * a dtree file `file_name + ".dtree"`.
            * (optionally) a variable file `file_name + ".exist"`,
                specifying that the variables in the file should be projected away.
                auxilliary variables are safe to project away.
                the resulting circuit will not be smooth if the exist file is provided!

        For miniC2D assumes that there is:
            * a cnf file `file_name`.
            * a vtree file `file_name + ".vtree"`.

        For d4 assumes that there is:
            * a cnf file `file_name`.

        For sharpsat-td assumes that there is: 
            * a cnf file `file_name`, which has knowledge compilation weights (see `CNF.write_kc_cnf()` for how that should look).

        An D/Vtree can be generated by using the functionality of aspmc.compile.dtree / aspmc.compile.vtree.

        Args:
            file_name (:obj:`string`): Path to the CNF and files containing the parameters for the knowledge compiler.
            knowledge_compiler (:obj:`string`, optional): The knowledge compiler to use. Defaults to `sharpsat-td`.
        Returns:
            None
        """        
        # compute the available memory to set the cache size
        available_memory = max(psutil.virtual_memory().available//1024**2 - 125, 1000)

        my_signals.tempfiles.add(file_name + '.nnf')
        logger.debug("Knowledge compiler output:")
        if knowledge_compiler == "c2d":
            if os.path.isfile(f"{file_name}.exist"):
                p = subprocess.Popen([os.path.join(src_path, "c2d/bin/c2d_linux"), "-reduce", "-in", file_name, "-dt_in", file_name + ".dtree", "-cache_size", str(available_memory), "-exist", file_name + ".exist"], stdout=subprocess.PIPE)
            else:
                p = subprocess.Popen([os.path.join(src_path, "c2d/bin/c2d_linux"), "-smooth_all", "-reduce", "-in", file_name, "-dt_in", file_name + ".dtree", "-cache_size", str(available_memory)], stdout=subprocess.PIPE)
        elif knowledge_compiler == "miniC2D":            
            p = subprocess.Popen([os.path.join(src_path, "miniC2D/bin/linux/miniC2D"), "-c", file_name, "-v", file_name + ".vtree", "-s" , str(available_memory)], stdout=subprocess.PIPE)
        elif knowledge_compiler == "sharpsat-td":
            decot = float(config.config["decot"])
            decot = max(decot, 0.1)
            p = subprocess.Popen(["./sharpSAT", "-dDNNF", "-decot", str(decot), "-decow", "100", "-tmpdir", "/tmp/", "-cs", str(available_memory//2), file_name, "-dDNNF_out", file_name + ".nnf"], cwd=os.path.join(src_path, "sharpsat-td/bin/"), stdout=subprocess.PIPE)
        elif knowledge_compiler == "d4":
            p = subprocess.Popen([os.path.join(src_path, "d4/d4_static"), file_name, "-dDNNF", f"-out={file_name}.nnf", "-smooth"], stdout=subprocess.PIPE)
        
        for line in iter(p.stdout.readline, b''):
            line = line.decode()
            logger.debug(line[:-1])
        p.wait()
        p.stdout.close()

        if p.returncode != 0:
            logger.error(f"Knowledge compilation failed with exit code {p.returncode}.")
            exit(-1) 

    def solve_compilation_single(self):
        """Compiles an AMC instance over a single semiring and performs the algebraic model counting over the compiled circuit.

        Currently supports c2d, miniC2D, d4 , sharpsat-td-live and sharpsat-td as knowledge compilers. 
        Generates a D/Vtree from a tree decomposition of the cnf for all knowledge compilers except d4.
        How the tree decomposition is generated and which knowledge compiler is used is configured in aspmc.config.

        Returns:
            object: The value of the AMC instance.
        """
        start = time.time()
        cnf_fd, cnf_tmp = tempfile.mkstemp()
        my_signals.tempfiles.add(cnf_tmp)
        # sharpsat-td-live is a special case since it does not fall into the `first compile then evaluate category`
        if config.config["knowledge_compiler"] == "sharpsat-td-live":
            with os.fdopen(cnf_fd, 'wb') as cnf_file:
                self.write_kc_cnf(cnf_file)
            decot = float(config.config["decot"])
            decot = max(decot, 0.1)
            # compute the available memory to set the cache size
            available_memory = max(psutil.virtual_memory().available//1024**2 - 125, 1000)
            p = subprocess.Popen(["./sharpSAT", "-instant", "-dDNNF", "-decot", str(decot), "-decow", "100", "-tmpdir", "/tmp/", "-cs", str(available_memory//2), cnf_tmp], cwd=os.path.join(src_path, "sharpsat-td/bin/"), stdout=subprocess.PIPE)
            weights, zero, one, dtype = self.get_weights()
            results = Circuit.live_parse_wmc(p.stdout, weights, zero = zero, one = one, dtype = dtype)
            p.wait()
            p.stdout.close()        
            if p.returncode != 0:
                logger.error(f"Knowledge compilation failed with exit code {p.exitcode}.")
                exit(-1) 
            end = time.time()
            logger.info(f"Counting & Compilation time:  {end - start}")
            os.remove(cnf_tmp)
            my_signals.tempfiles.remove(cnf_tmp)
            return results
        
        # prepare everything for the compilation
        v3 = None
        if config.config["knowledge_compiler"] == "c2d":
            with os.fdopen(cnf_fd, 'wb') as cnf_file:
                self.to_stream(cnf_file)
            d3 = dtree.TD_dtree(self, solver = config.config["decos"], timeout = config.config["decot"])
            d3.write(cnf_tmp + '.dtree')
            my_signals.tempfiles.add(cnf_tmp + '.dtree')
            end = time.time()
            logger.info(f"Dtree time:               {end - start}")
            if len(self.semirings) == 1 and self.semirings[0].__name__ == "aspmc.semirings.probabilistic":
                # careful: with other semirings this does not generally work, since the "-smooth_all" option of c2d is necessary in general then
                # this is a problem since -smooth_all also adds the variables that were existentially quantified away again
                with open(cnf_tmp + '.exist', 'w') as exist_file:
                    exist_file.write(f"{len(self.auxilliary)} {' '.join( str(v) for v in self.auxilliary )}")
                my_signals.tempfiles.add(cnf_tmp + '.exist')
        elif config.config["knowledge_compiler"] == "miniC2D":            
            with os.fdopen(cnf_fd, 'wb') as cnf_file:
                self.to_stream(cnf_file)
            v3 = vtree.TD_vtree(self, solver = config.config["decos"], timeout = config.config["decot"])
            v3.write(cnf_tmp + ".vtree")
            my_signals.tempfiles.add(cnf_tmp + '.vtree')
            end = time.time()
            logger.info(f"Vtree time:               {end - start}")
        elif config.config["knowledge_compiler"] == "sharpsat-td":
            with os.fdopen(cnf_fd, 'wb') as cnf_file:
                self.write_kc_cnf(cnf_file)
        elif config.config["knowledge_compiler"] == "d4":
            with os.fdopen(cnf_fd, 'wb') as cnf_file:
                self.to_stream(cnf_file)
                
        # perform the actual compilation
        start = time.time()
        CNF.compile_single(cnf_tmp, knowledge_compiler = config.config["knowledge_compiler"])
        end = time.time()
        logger.info(f"Compilation time:         {end - start}")
        # perform the counting on the circuit
        weights, zero, one, dtype = self.get_weights()
        start = time.time()
        results = Circuit.parse_wmc(cnf_tmp + ".nnf", weights, zero = zero, one = one, dtype = dtype, solver = config.config["knowledge_compiler"], vtree = v3)
        end = time.time()
        logger.info(f"Counting time:            {end - start}")
        
        # remove the temporary files
        os.remove(cnf_tmp)
        my_signals.tempfiles.remove(cnf_tmp)
        os.remove(cnf_tmp+".nnf")
        my_signals.tempfiles.remove(cnf_tmp + '.nnf')
        if config.config["knowledge_compiler"] == "c2d":
            os.remove(cnf_tmp + ".dtree")
            my_signals.tempfiles.remove(cnf_tmp + '.dtree')
            if (cnf_tmp + ".exist") in my_signals.tempfiles:
                os.remove(cnf_tmp + ".exist")
                my_signals.tempfiles.remove(cnf_tmp + '.exist')
        elif config.config["knowledge_compiler"] == "miniC2D":
            os.remove(cnf_tmp + ".vtree")
            my_signals.tempfiles.remove(cnf_tmp + '.vtree')
        return results


    @staticmethod
    def compile_nested(file_name, knowledge_compiler = "c2d"):        
        """Compiles a CNF into an X/D-constrained circuit. The output circuit is in the file `file_name + ".nnf"`.

        Currently supports c2d (and miniC2D for only one level of nesting) as knowledge compilers. 

        For c2d assumes that there is:
            * a cnf file `file_name` 
            * a dtree file `file_name + ".dtree"`
            * a force file `file_name + ".force"`

        For miniC2D assumes that there is:
            * a cnf file `file_name` 
            * a vtree file `file_name + ".vtree"`
            
        An X/D-constrained D/Vtree can be generated by using the functionality of aspmc.compile.constrained_compile.

        Args:
            file_name (:obj:`string`): Path to the CNF and files containing the parameters for the knowledge compiler.
            knowledge_compiler (:obj:`string`, optional): The knowledge compiler to use. Defaults to `c2d`.
        Returns:
            None
        """
        # compute the available memory to set the cache size
        available_memory = max(psutil.virtual_memory().available//1024**2 - 125, 1000)

        my_signals.tempfiles.add(file_name + '.nnf')
        logger.debug("Knowledge compiler output:")

        if knowledge_compiler == "c2d":
            p = subprocess.Popen([os.path.join(src_path, "c2d/bin/c2d_linux"), "-cache_size", str(available_memory), "-keep_trivial_cls", "-smooth_all", "-in", file_name, "-dt_in", file_name + ".dtree", "-force", file_name + ".force"], stdout=subprocess.PIPE)
        elif knowledge_compiler == "miniC2D":
            p = subprocess.Popen([os.path.join(src_path, "miniC2D/bin/linux/miniC2D"), "-c", file_name, "-v", file_name + ".vtree", "-s" , str(available_memory)], stdout=subprocess.PIPE)
        else:
            logger.error(f"Knowledge compiler {config.config['knowledge_compiler']} does not support X/D-constrained compilation")
            exit(-1)

        for line in iter(p.stdout.readline, b''):
            line = line.decode()
            logger.debug(line[:-1])
        p.wait()
        p.stdout.close()

        if p.returncode != 0:
            logger.error(f"Knowledge compilation failed with exit code {p.exitcode}.")
            exit(-1) 

    def solve_compilation_nested(self):
        """Compiles a kAMC instance over a k semirings into an X/D-constrained circuit and performs the algebraic model counting over the compiled circuit.

        Currently supports c2d (and miniC2D for 2AMC) as knowledge compilers. 
        Generates an X/D-constrained D/Vtree by using the function in aspmc.compile.constrained_compile.
        How the tree decompositions are generated and which knowledge compiler is used is configured in aspmc.config.

        Returns:
            object: The value of the kAMC instance.
        """
        start = time.time()
        cnf_fd, cnf_tmp = tempfile.mkstemp()
        my_signals.tempfiles.add(cnf_tmp)
        if config.config["knowledge_compiler"] == "c2d":
            (force_vars, d3) = concom.tree_from_cnf(self, tree_type = dtree.Dtree)
            d3.write(cnf_tmp + ".dtree")
            my_signals.tempfiles.add(cnf_tmp + '.dtree')
            with os.fdopen(cnf_fd, 'wb') as cnf_file:
                self.to_stream(cnf_file)
            my_signals.tempfiles.add(cnf_tmp + '.force')
            with open(cnf_tmp + ".force", 'w') as force_out:
                force_out.write(f"{len(force_vars)} {' '.join([ str(v) for v in force_vars ])}")
            end = time.time()
            logger.info(f"Dtree time:               {end - start}")
        elif config.config["knowledge_compiler"] == "miniC2D":
            assert(len(self.semirings) == 2)
            with os.fdopen(cnf_fd, 'wb') as cnf_file:
                self.to_stream(cnf_file)
            (_, v3) = concom.tree_from_cnf(self, tree_type=vtree.Vtree)
            v3.write(cnf_tmp + ".vtree")
            my_signals.tempfiles.add(cnf_tmp + '.vtree')
            end = time.time()
            logger.info(f"Vtree time:               {end - start}")
        else:
            logger.error(f"Knowledge compiler {config.config['knowledge_compiler']} does not support X/D-constrained compilation")
            exit(-1)
        # perform the compilation
        start = time.time()
        CNF.compile_nested(cnf_tmp, knowledge_compiler = config.config["knowledge_compiler"])
        end = time.time()
        logger.info(f"Compilation time:         {end - start}")
        # prepare the inputs
        start = time.time()
        P = set(self.quantified[0])
        weights = []
        for i in range(len(self.weights)):
            weights.append(self.weights[to_dimacs(i)])
        if config.config["knowledge_compiler"] == "c2d":
            circ = ConstrainedDDNNF
        else:
            circ = ConstrainedSDD(path = None, v3 = v3)
        end = time.time()
        logger.info(f"Preparation time:         {end - start}")
        start = time.time()
        results = circ.parse_wmc(cnf_tmp + '.nnf', weights, self)
        end = time.time()
        logger.info(f"Counting time:            {end - start}")
        # clean up the files
        os.remove(cnf_tmp)
        my_signals.tempfiles.remove(cnf_tmp)
        os.remove(cnf_tmp+".nnf")
        my_signals.tempfiles.remove(cnf_tmp + '.nnf')
        if config.config["knowledge_compiler"] == "c2d":
            os.remove(cnf_tmp + ".dtree")
            my_signals.tempfiles.remove(cnf_tmp + '.dtree')
            os.remove(cnf_tmp + ".force")
            my_signals.tempfiles.remove(cnf_tmp + '.force')
        else:
            os.remove(cnf_tmp + ".vtree")
            my_signals.tempfiles.remove(cnf_tmp + '.vtree')
        return results

    def preprocessing(self):
        start = time.time()
        if len(self.semirings) == 1 and self.semirings[0].is_idempotent(): # TODO make sure is_idempotent() is always implemented
            mode = "idemp"
        else:
            mode = "general"
        
        (cnf_file_fd, cnf_file_tmp) = tempfile.mkstemp()
        my_signals.tempfiles.add(cnf_file_tmp)

        with os.fdopen(cnf_file_fd, mode = 'w') as cnf_file:
            cnf_file.write(str(self)) 
        
        q = subprocess.Popen([os.path.join(src_path, "preprocessor/bin/sharpSAT"), "-m", mode, "-t", "FPVEG", cnf_file_tmp], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output, err = q.communicate()
        
        self.clauses = [] 
        cnf_reached = False
        for line in output.decode().split('\n'):
            line = line.split()
            if len(line) > 0 and not line[0] == 'c':
                if line[0] == 'p' and line[1] == "cnf":
                    cnf_reached = True
                    self.nr_vars = int(line[2]) # TODO: instead check if value has changed (which should not happen)
                elif cnf_reached:
                    line = [int(l) for l in line]
                    self.clauses.append(line[:-1])
        end = time.time()
        os.remove(cnf_file_tmp)
        my_signals.tempfiles.remove(cnf_file_tmp)
        logger.info(f"Preprocessing time:       {end - start}")

    def evaluate(self, strategy = "flexible", preprocessing = False):
        """Evaluates an AMC instance by using the given strategy `strategy`.
        
        The strategy can be one of 
            * `"flexible"`: do what is assumed best. 
                Currently, uses maxsat for idempotent semirings and knowledge compilation otherwise.
            * `"compilation"`: compile regardless of the properties of the semiring.

        Calls `solve_compilation` or `solve_maxsat` depending on the problem and strategy.
        How the tree decompositions are generated and which knowledge compiler/maxsat solver is used is configured in aspmc.config.

        Args:
            strategy (:obj:`string`, optional): Which strategy to use for evaluation. Default is `"flexible"`.
        Returns:
            object: The value of the AMC instance.
        """
        if strategy == "flexible":
            if len(self.semirings) == 1 and self.semirings[0].is_idempotent():
                return self.solve_maxsat()
            elif len(self.semirings) == 1 and config.config["knowledge_compiler"] == "sharpsat-td"\
                 and self.semirings[0].__name__ == "aspmc.semirings.probabilistic":
                return self.solve_wmc()
            elif len(self.semirings) == 0 and config.config["knowledge_compiler"] == "sharpsat-td":
                return self.solve_mc()
            else:
                return self.solve_compilation(preprocessing = preprocessing)
        elif strategy == "compilation":
            return self.solve_compilation(preprocessing = preprocessing)
        else: 
            logger.error(f"Unknown evaluation strategy {strategy}.")
            exit(-1)

    def solve_compilation(self, preprocessing = False):   
        """Compiles an AMC instance and performs the algebraic model counting over the compiled circuit.

        Calls `compile_single` or `compile_nested`.
        How the tree decompositions are generated and which knowledge compiler is used is configured in aspmc.config.

        Returns:
            object: The value of the AMC instance.
        """
        if preprocessing:
            logger.info("Preprocessing enabled")
            self.preprocessing()
            logger.info("------------------------------------------------------------")
        else:
            logger.info("Preprocessing disabled")
        results = self.evaluate_trivial()
        if results is not None:
            return results
        logger.info("   Stats Compilation")
        logger.info("------------------------------------------------------------")
        if len(self.semirings) <= 1:
            results = self.solve_compilation_single()
        else:
            results = self.solve_compilation_nested()
        logger.info("------------------------------------------------------------")
        return results

    def solve_maxsat(self):
        cnf_fd, cnf_tmp = tempfile.mkstemp()
        my_signals.tempfiles.add(cnf_tmp)
        logger.debug(f"    MaxSAT CNF file: {cnf_tmp}")
        # first we check whether this is actually a MaxSAT instance or whether it is just a SAT instance in disguise
        import math
        real_weights = { l : w[0].value for l,w in self.weights.items() }
        if self.semirings[0].__name__ == "aspmc.semirings.maxtimes":
            # we first need to convert into maxplus weights
            real_weights = { l : math.log(w) if w > 0 else float("-inf") for l,w in real_weights.items() }
        elif self.semirings[0].__name__ == "aspmc.semirings.minplus":
            # we first need to convert into maxplus weights
            real_weights = { l : -w if w != float("inf") else float("-inf") for l,w in real_weights.items() }
        elif self.semirings[0].__name__ != "aspmc.semirings.maxplus":
            logger.error(f"MaxSAT evaluation is currently not supported for semiring {self.semirings[0].__name__}")
            exit(-1)
        # sort out variables that are irrelevant
        real_weights = { l : w for l,w in real_weights.items() if w != real_weights[-l]}
        # handle hard constraints from literals with weight -inf and keep the rest
        negated_units = [ l for l,w in real_weights.items() if w == float('-inf')]
        real_weights = { l : w for l,w in real_weights.items() if w != float('-inf') and abs(w) > 0}


        (weights, zero, one, dtype) = self.get_weights()
        first_shape = (np.shape(weights[0])[0], ) + np.shape(one)
        if len(real_weights) == 0:
            # this is a SAT instance!
            with os.fdopen(cnf_fd, mode='wb') as cnf_out:
                # create result file
                res_fd, res_tmp = tempfile.mkstemp()
                my_signals.tempfiles.add(res_tmp)
                # write the cnf with the additional negated unit literals
                cnf_out.write(f"p cnf {self.nr_vars} {len(self.clauses) + len(negated_units)}\n".encode())
                for c in self.clauses:
                    cnf_out.write(f"{' '.join([str(l) for l in c])} 0\n".encode())
                for lit in negated_units:
                    cnf_out.write(f"{-lit} 0\n".encode())
                # solve
                p = subprocess.Popen([os.path.join(src_path, "minisat-definitions/bin/minisat"), cnf_tmp, res_tmp], stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds = True)
                p.wait()
                output = p.stdout.read().decode().split("\n")
                for line in output:
                    line = line.split()
                    if len(line) == 0:
                        continue
                    if line[0] == 'UNSATISFIABLE':
                        weight = np.empty(first_shape, dtype=dtype)
                        weight[:] = zero
                        solution = list(range(1,self.nr_vars + 1))
                        p.stdout.close()
                    elif line[0] == 'SATISFIABLE':
                        with os.fdopen(res_fd, mode='r') as result_file:
                            solution = result_file.read().split('\n')[1]
                        solution = [ int(v) for v in solution.split(' ') if v != '' ]
                        weight = np.empty(first_shape, dtype=dtype)
                        weight[:] = one
                        for lit in solution:
                            weight *= weights[to_pos(lit)]
                        p.stdout.close()
                os.remove(res_tmp)
                my_signals.tempfiles.remove(res_tmp)
        else:
            logger.info("   Stats MaxSAT")
            logger.info("------------------------------------------------------------")
            start = time.time()
            with os.fdopen(cnf_fd, mode='wb') as cnf_out:
                self.write_maxsat_cnf(cnf_out)
            p = subprocess.Popen([os.path.join(src_path, "UWrMaxSAT/uwrmaxsat/build/release/bin/uwrmaxsat"), "-no-bin", "-no-sat", "-m", "-bm", "-maxpre-time=10", cnf_tmp], stdout=subprocess.PIPE, close_fds = True)#, stderr=subprocess.PIPE)
            solution = None
            while p.poll() is None or solution is None:
                line = p.stdout.readline().decode()
                if len(line) == 0:
                    continue
                if line.startswith("s"):
                    if line[2:] == "OPTIMUM FOUND":
                        continue
                    elif line[2:] == "UNKNOWN":
                        raise SolvingError("MaxSAT solver returned UNKNOWN")
                    elif line[2:] == "SATISFIABLE":
                        raise SolvingError("MaxSAT solver returned SATISFIABLE. Probably it was interrupted during execution")
                    elif line[2:] == "UNSATISFIABLE":
                        weight = np.empty(first_shape, dtype=dtype)
                        weight[:] = zero
                        solution = list(range(1,self.nr_vars + 1))
                elif line[0] == 'v':
                    bitset = line[2:-1]
                    solution = [ i if bitset[i-1] == '1' else -i for i in range(1,self.nr_vars + 1)]
                    weight = np.empty(first_shape, dtype=dtype)
                    weight[:] = one
                    for lit in solution:
                        weight *= weights[to_pos(lit)]
            p.stdout.close()
            if solution is None:
                raise SolvingError("MaxSAT solver did not print an assignment!")
            
            logger.info(f"Solving time:         {time.time() - start}")
            logger.info("------------------------------------------------------------")
        os.remove(cnf_tmp)
        my_signals.tempfiles.add(cnf_tmp)
        return weight


    def solve_wmc(self):
        _, cnf_tmp = tempfile.mkstemp()
        my_signals.tempfiles.add(cnf_tmp)
        logger.debug(f"    WCNF file: {cnf_tmp}")
        self.to_file(cnf_tmp, extras=True)
        logger.info("   Stats Model Counter")
        logger.info("------------------------------------------------------------")
        start = time.time()
        decot = float(config.config["decot"])
        decot = max(decot, 0.1)
        # compute the available memory to set the cache size
        available_memory = max(psutil.virtual_memory().available//1024**2 - 125, 1000)
        first = None
        for weight in self.weights.values():
            first = weight
            break
        p = subprocess.Popen(["./sharpSAT", "-MWD", str(len(first)), "-decot", str(decot), "-decow", "100", "-tmpdir", "/tmp/", "-cs", str(available_memory//2), cnf_tmp], cwd=os.path.join(src_path, "sharpsat-td/bin/"), stdout=subprocess.PIPE)
        result = None
        logger.debug("Solver output:")
        for line in iter(p.stdout.readline, b''):
            line = line.decode()
            if line.startswith("c s exact arb float "):
                result = np.array([ float(part) for part in line[len("c s exact arb float "):-1].split(";") ])
            logger.debug(line[:-1])
        p.wait()
        p.stdout.close()
        if result is None:
            raise SolvingError("Model Counter did not print a solution!")
        
        logger.info(f"Counting time:         {time.time() - start}")
        logger.info("------------------------------------------------------------")
        os.remove(cnf_tmp)
        my_signals.tempfiles.add(cnf_tmp)
        return result

    def solve_mc(self):
        _, cnf_tmp = tempfile.mkstemp()
        my_signals.tempfiles.add(cnf_tmp)
        logger.debug(f"    CNF file: {cnf_tmp}")
        self.to_file(cnf_tmp, extras=False)
        logger.info("   Stats Model Counter")
        logger.info("------------------------------------------------------------")
        start = time.time()
        decot = float(config.config["decot"])
        decot = max(decot, 0.1)
        # compute the available memory to set the cache size
        available_memory = max(psutil.virtual_memory().available//1024**2 - 125, 1000)
        p = subprocess.Popen(["./sharpSAT", "-decot", str(decot), "-decow", "100", "-tmpdir", "/tmp/", "-cs", str(available_memory//2), cnf_tmp], cwd=os.path.join(src_path, "sharpsat-td/bin/"), stdout=subprocess.PIPE)
        result = None
        logger.debug("Solver output:")
        for line in iter(p.stdout.readline, b''):
            line = line.decode()
            if line.startswith("c s exact arb int "):
                result = np.array([ int(line[len("c s exact arb int "):-1]) ])
            logger.debug(line[:-1])
        p.wait()
        p.stdout.close()
        if result is None:
            raise SolvingError("Model Counter did not print a solution!")
        
        logger.info(f"Counting time:         {time.time() - start}")
        logger.info("------------------------------------------------------------")
        os.remove(cnf_tmp)
        my_signals.tempfiles.add(cnf_tmp)
        return result



import networkx as nx
import logging

import aspmc.graph.treedecomposition as treedecomposition

import aspmc.compile.vtree as vtree
import aspmc.compile.dtree as dtree

import aspmc.config as config

logger = logging.getLogger("aspmc")

def compute_separator(graph, P, D, R):
    if len(P) == 0 or len(R) == 0:
        return []
    flow_graph = nx.DiGraph()
    N = graph.number_of_nodes()
    # 1 to N are nodes for incoming edges
    # N + 1 to 2*N are nodes for outgoing edges
    # the edges between i, i + N for i in P or D are the 
    # only ones that can be cut
    for i in P:
        flow_graph.add_edge(i, i + N, capacity = 1.0)
    for i in D:
        flow_graph.add_edge(i, i + N, capacity = 1.0)
    for i in R:
        # these are edges with infinite capacity
        flow_graph.add_edge(i, i + N)
        
    for u,v in graph.edges:
        flow_graph.add_edge(u + N, v)
        flow_graph.add_edge(v + N, u)
    
    # add global source and sink
    source = 2*N + 1
    sink = 2*N + 2
    flow_graph.add_node(source)
    flow_graph.add_node(sink)
    for i in R:
        flow_graph.add_edge(source, i)
    for i in P:
        flow_graph.add_edge(i + N, sink)
        
    cut_value, partition = nx.minimum_cut(flow_graph, source, sink)
    reachable, non_reachable = partition
    cutset = set()
    for u, nbrs in ((n, flow_graph[n]) for n in reachable):
        cutset.update((u, v) for v in nbrs if v in non_reachable)
    return [ u for u, _ in cutset ]

def TD_to_tree(cnf, td, done = None, tree_type = dtree.Dtree):
    if tree_type.__name__ is dtree.Dtree.__name__:
        return dtree.TD_to_dtree(cnf, td, done = done)
    elif tree_type.__name__ == vtree.Vtree.__name__:
        return vtree.TD_to_vtree(td)
    else:
        logger.error(f"Unknown tree type {tree_type}")
        exit(-1)

def from_order(cnf, order, done = None, tree_type = dtree.Dtree):
    if tree_type.__name__ is dtree.Dtree.__name__:
        return dtree.from_order(cnf, order, done = done)
    elif tree_type.__name__ == vtree.Vtree.__name__:
        return vtree.from_order(order)
    else:
        logger.error(f"Unknown tree type {tree_type}")
        exit(-1)


def tree_from_cnf(cnf, tree_type = dtree.Dtree):    
    """Constructs an X/D-constrained Vtree or Dtree for the given cnf.

    Does this by: 

    * getting the atoms `X` from `cnf.quanfied[0]`. So the atoms that are quantified over the first semiring.
    * getting the atoms `D` that are defined by `X` w.r.t. the cnf.
    * getting a separator `S` between the atoms in `X` and the ones neither in `X` nor `D` using atoms from `X` or `D`.
    * the constructed X/D-constrained D/Vtree then ensures that all the atoms in `S` are decided first,
        meaning that all the atoms in `X` can be decided before/independently of the atoms that are not in `X` or `D`.
    * if there are more than two semirings, this is repeated recursively.

    Args:
        cnf (:obj:`aspmc.compile.cnf.CNF`): The extended cnf for which a tree should be constructed. Must have exactly two semirings.
        tree_type (:obj:`type`, optional): The type of tree to construct. 
            Must be one of `aspmc.compile.dtree.Dtree` and `aspmc.compile.vtree.Vtree`.
            Defaults to `aspmc.compile.dtree.Dtree`.
    Returns:
        (iterable, aspmc.graph.bintree.bintree): 

        The vertices in the separator(s) `S` that was (were) computed.
        
        The root of the D/Vtree that was constructed.
    """
    P = set()
    D = set()
    R = set(range(1,cnf.nr_vars + 1))
    graph = cnf.primal_graph()
    separators = []
    for i in range(len(cnf.quantified) - 1):
        P.update(cnf.quantified[i])
        R.difference_update(cnf.quantified[i])
        
        if config.config["constrained"] == "XD":
            D = cnf.get_defined(P)
            R.difference_update(D)
        
        # split the whole graph into two graphs that only contain nodes from P U D or R U D 
        separator = compute_separator(graph, P, D, R)
        logger.debug(f"Size of the {i}-th separator: {len(separator)}")
        separators.append(separator)
        
        # remove all edges that use separator nodes from the graph
        graph.remove_nodes_from(separator)
        # readd the nodes for invariance reasons
        graph.add_nodes_from(separator)
    
    graph = cnf.primal_graph()
    tree = construct_tree(cnf, graph, separators, tree_type = tree_type)
    
    # gather all variables that occur in a separator 
    all_separators = set()
    for separator in separators:
        all_separators.update(separator)
    return [all_separators, tree]

def construct_tree(cnf, graph, separators, tree_type = dtree.Dtree):
    done = None
    if tree_type.__name__ is dtree.Dtree.__name__: 
        # remember which clauses have been taken care of already
        done = [ False for _ in range(len(cnf.clauses)) ]
        
    left_trees = []
    working_graph = graph.copy()
    overall_l_components = set()
    previous_separators = set()
    for level in range(len(cnf.quantified) - 1):
        separator = separators[level]
        previous_separators.update(separator)
        P = cnf.quantified[level]    
        # if the separator is empty we know all variables are defined or P is empty
        # either way, we have no restrictions here
        if len(separator) == 0:
            left_trees.append(None)
            continue
        
        # remove the nodes from the graph
        working_graph.remove_nodes_from(separator)

        # build the graphs that contain only nodes from P U D or R U D respectively
        l_components = set()
        for component in nx.connected_components(working_graph):
            if P & component:
                l_components.update(component)

        overall_l_components.update(l_components)

        if len(l_components) == 0:
            # we used P as the separator
            left_trees.append(None)
        else:
            # also remove the left part from the working graph
            working_graph.remove_nodes_from(l_components)
            # we found a better separator than P
            l_components.update(previous_separators)
            l_graph = graph.subgraph(l_components).copy()
            separator = list(separator)
            clique = sum([ [ (separator[a], separator[b]) for a in range(b + 1, len(separator)) ] for b in range(len(separator)) ], [])
            l_graph.add_edges_from(clique)
            l_td = treedecomposition.from_graph(l_graph, solver = config.config["decos"], timeout = config.config["decot"])
            logger.info(f"X/D-Constrained Tree Decomposition #bags: {l_td.bags} treewidth: {l_td.width} #vertices: {l_td.vertices}")
            l_root = l_td.find_containing(separator)
            l_td.set_root(l_root)
            if tree_type.__name__ == vtree.Vtree.__name__:
                l_td.remove(separator)
            l_tree = TD_to_tree(cnf, l_td, done = done, tree_type = tree_type)
            left_trees.append(l_tree)
            
    right_graph = graph.copy()
    right_graph.remove_nodes_from(overall_l_components)
    for level in range(len(cnf.quantified) - 1):
        separator = separators[level]
        separator = list(separator)
        clique = sum([ [ (separator[a], separator[b]) for a in range(b + 1, len(separator)) ] for b in range(len(separator)) ], [])
        right_graph.add_edges_from(clique)
    right_td = treedecomposition.from_graph(right_graph, solver = config.config["decos"], timeout = config.config["decot"])
    logger.info(f"X/D-Constrained Tree Decomposition #bags: {right_td.bags} treewidth: {right_td.width} #vertices: {right_td.vertices}")
    
    r_root = right_td.find_containing(separators[0])
    assert(r_root is not None)
    right_td.set_root(r_root)
    
    if tree_type.__name__ == vtree.Vtree.__name__:
        for level in range(len(cnf.quantified) - 1):
            separator = separators[level]
            right_td.remove(separator)
            
    parent = TD_to_tree(cnf, right_td, done = done, tree_type = tree_type)
    # all materials are in place
    # construct the complete tree by bottom up stitching
    for level in range(len(separators) - 1, -1, -1):
        separator = separators[level]
        
        if len(separator) == 0:
            # we do not have to do anything
            continue
        
        # first combine right and left
        if parent is None:
            parent = left_trees[level]
        elif left_trees[level] is None:
            pass
        else:
            new_parent = tree_type()
            new_parent.left = left_trees[level]
            new_parent.right = parent
            parent = new_parent
            
        sep_tree = from_order(cnf, separator, done = done, tree_type = tree_type)
        
        # find out where to put the parent
        last = None
        cur = sep_tree
        while not cur.right is None:
            last = cur
            cur = cur.right
        grandparent = tree_type()
        if not last is None:
            last.right = grandparent
        else:
            sep_tree = grandparent
        grandparent.left = cur
        grandparent.right = parent
        
        parent = sep_tree
        
    if tree_type.__name__ is dtree.Dtree.__name__ and not all(done):
        logger.error("Not all clauses are in the dtree.")
        exit(-1)
        
    return parent
            


"""
Program module providing the two algebraic progam class.
"""
import numpy as np

import logging
logger = logging.getLogger("aspmc")

from aspmc.programs.program import Program

from aspmc.util import *

class TwoAlgebraicProgram(Program):    
    """A class for programs with weights over two semirings. 

    This is supposed to be an abstract class in the sense that it should used and instantiated
    by subclasses, rather than parsed itself.

    Overrides the `get_weights`, `_finalize_cnf` and `get_queries` methods appropriately.

    Args:
        clingo_control (:obj:`Control`): The clingo control object having the rules of the spanning program.
        first_semiring (:obj:`module`): The first semiring module.
        second_semiring (:obj:`module`): The second semiring module.
        first_weights (:obj:`dict`): The weights over the first semiring. 
            Needs to map pairs `(name, phase)`, where `name` is the name of an atom 
            and `phase` is its polarity (`True` for true), to their weight.
        second_weights (:obj:`dict`): The weights over the second semiring. 
            Needs to map pairs `(name, phase)`, where `name` is the name of an atom 
            and `phase` is its polarity (`True` for true), to their weight.
        transform (:obj:`string`): A string representation of the function that transforms a value from the first
            into a value from the second semiring.
        queries (:obj:`list`): A list of names of atoms that should be queried. 
            Specify the empty list for an overall weight query.

    Attributes:
        first_semiring (:obj:`module`): The first semiring module.
        second_semiring (:obj:`module`): The second semiring module.
        first_weights (:obj:`dict`): The weights over the first semiring. 
            Needs to map pairs `(name, phase)`, where `name` is the name of an atom 
            and `phase` is its polarity (`True` for true), to their weight.
        second_weights (:obj:`dict`): The weights over the second semiring. 
            Needs to map pairs `(name, phase)`, where `name` is the name of an atom 
            and `phase` is its polarity (`True` for true), to their weight.
        transform (:obj:`string`): A string representation of the function that transforms a value from the first
            into a value from the second semiring.
        queries (:obj:`list`): A list of names of atoms that should be queried. 
            Specify the empty list for an overall weight query.
    """
    def __init__(self, clingo_control, first_semiring, second_semiring, first_weights, second_weights, transform, queries):
        logger.warning("TwoAlgebraicProgram may be removed in favor of the new NestedAlgebraicProgram in the future")
        self.first_semiring = first_semiring
        self.second_semiring = second_semiring
        self.first_weights = first_weights
        self.second_weights = second_weights
        self.transform = transform
        self.queries = queries
        Program.__init__(self, clingo_control = clingo_control)

    def _finalize_cnf(self):
        weight_list = self.get_weights()
        varMap = { name : var for var, name in self._nameMap.items() }
        first = set([ int(str(varMap[name])) for (name, _) in self.first_weights ])
        second = set([ int(str(varMap[name])) for (name, _) in self.second_weights ])
        self._cnf.quantified = [first, second]
        for v in range(self._cnf.nr_vars*2):
            self._cnf.weights[to_dimacs(v)] = weight_list[v]
        self._cnf.transforms = [ self.transform ]
        self._cnf.semirings = [ self.first_semiring, self.second_semiring ]
        # TODO figure out which variables we can mark as auxilliary

    def get_weights(self):
        query_cnt = max(len(self.get_queries()), 1)
        first_shape = (query_cnt, ) + np.shape(self.first_semiring.one())
        second_shape = (query_cnt, ) + np.shape(self.second_semiring.one())
        varMap = { name : var for var, name in self._nameMap.items() }
        weight_list = [ np.empty(second_shape, dtype=self.second_semiring.dtype) for _ in range(self._cnf.nr_vars*2) ]
        for i in range(self._cnf.nr_vars*2):
            weight_list[i][:] = self.second_semiring.one()
        for (name, phase) in self.first_weights:
            idx = to_pos(varMap[name])
            if not phase:
                idx = neg(idx)
            weight_list[idx] = np.empty(first_shape, dtype=self.first_semiring.dtype)
            weight_list[idx][:] = self.first_weights[(name, phase)]
        for (name, phase) in self.second_weights:            
            idx = to_pos(varMap[name])
            if not phase:
                idx = neg(idx)
            weight_list[idx] = np.empty(second_shape, dtype=self.second_semiring.dtype)
            weight_list[idx][:] = self.second_weights[(name, phase)]
        for i, query in enumerate(self.get_queries()):
            if (query, True) in self.first_weights:
                weight_list[neg(to_pos(varMap[query]))][i] = self.first_semiring.zero()
            else:
                weight_list[neg(to_pos(varMap[query]))][i] = self.second_semiring.zero()
        return weight_list

    def get_queries(self):
        return self.queries



import networkx as nx

class Hypergraph(object):
    """A class for Hypergraphs, i.e. graphs that have hyperedges, 
    meaning that they connect an arbitrary subset of vertices with one another.

    Attributes:
        nodes (:obj:`set`): The nodes/vertices of the hypergraph.
        edges (:obj:`list`): The (hyper)edges of the hypergraph.
    """
    def __init__(self):
        self.nodes = set()
        self.edges = []
    
    def add_node(self, node):
        """Add a single node/vertex to the hypergraph.
        Args:
            node (:obj:`int`): The id of the node to add.

        Returns:
            None
        """
        self.nodes.add(node)
    
    def add_nodes_from(self, nodes):        
        """Add a nodes/vertices from an iterable to the hypergraph.
        Args:
            nodes (:obj:`iterable`): The ids (integers!) of the node to add.

        Returns:
            None
        """
        self.nodes.update(nodes)

    def add_edge(self, edge):        
        """Add a single (hyper)edge to the hypergraph.
        Args:
            edge (:obj:`iterable`): The iterable of all the nodes in the edge.

        Returns:
            None
        """
        self.edges.append(edge)
    
    def add_edges_from(self, edges):
        """Add a (hyper)edges from an iterable to the hypergraph.
        Args:
            edges (:obj:`iterable`): The iterable of iterables all the nodes in the edge.

        Returns:
            None
        """
        self.edges += list(edges)

    def to_graph(self):
        """Convert the hypergraph to an `nx.Graph` by replacing hyperedges with cliques.

        Returns:
            :obj:`nx.Graph`: The graph corresponding to this hypergraph.
        """
        graph = nx.Graph()
        graph.add_nodes_from(self.nodes)
        for edge in self.edges:
            graph.add_edges_from(sum([ [ (v, vp) for v in edge if v != vp ] for vp in edge ], []))
        return graph


"""
Dtree module providing methods to construct, visualize and generally work with dtrees.
"""

import aspmc.graph.treedecomposition as treedecomposition
from aspmc.graph.bintree import bintree

import math

class Dtree(bintree):
    """A Dtree class that manages dtrees. 

    These can be used as the input to c2d to (loosely) specify in which order atoms are decided.
    Is a subclass of aspmc.graph.bintree.bintree and therefore has some utility functions, like writing to file, 
    counting the leaves or iterating over the nodes of the tree.
    """

    def __str__(self):
        self.index()
        res = ""
        res += f"dtree {2*self.leaf_count() - 1}\n"
        cur_max = -1
        def get_next():
            nonlocal cur_max
            cur_max += 1
            return cur_max
        for node in self:
            if node.val is not None:
                node.w_idx = get_next()
                res += f"L {node.val}\n"
            else:
                node.w_idx = get_next()
                res += f"I {node.left.w_idx} {node.right.w_idx}\n"
        return res

def from_order(cnf, order, done = None):        
    """Constructs a (partial) dtree for a cnf from a list of atoms in the specified order.

    Modifies the cnf by adding trivial clauses that are used to ensure that the atoms in `order` are eliminated first.
        
    Args:
        cnf (:obj:`aspmc.compile.cnf.CNF`): The cnf to construct the partial dtree for.
        order (:obj:`list`): The list specifying which atoms should be eliminated and the order in which they should be eliminated.
        done (:obj:`list`, optional): A list of Booleans specifying which clauses of the cnf are already included in a dtree. 
        If given updates the list by setting the value of the added clauses to True.
        
    Returns:
        :obj:`Dtree`: The (right-linear) dtree that eliminates the atoms in `order` in the order they are in `order`.
    """
    root = Dtree()
    cur = root
    for v in order[:-1]:
        cnf.clauses.append([v, -v])
        cur.left = Dtree(val = len(cnf.clauses) - 1)
        cur.right = Dtree()
        cur = cur.right
    cnf.clauses.append([order[-1], -order[-1]])
    cur.val = len(cnf.clauses) - 1
    if done != None:
        done += [ True for _ in range(len(order))]
    return root

# def binary(dtrees):
#     if len(dtrees) == 0:
#         return None
#     if len(dtrees) == 1:
#         return dtrees[0]
#     root = Dtree()
#     half = 2**(math.ceil(math.log2(len(dtrees)))-1)
#     root.left = binary(dtrees[:half])
#     root.right = binary(dtrees[half:])
#     return root

# def cw(cnf, td):
#     from aspmc.compile import vtree
#     v3 = vtree.TD_to_vtree(td)
#     order = [ n.val for n in v3.preorder() if n.val is not None ]
#     print(order)
#     v3.draw()
#     idx = [ 0 for _ in range(len(order) + 1) ]
#     for i,v in enumerate(order):
#         idx[v] = i
#     cuts = [ 0 for _ in range(len(order)) ]
#     for c in cnf.clauses:
#         max_v = -1
#         min_v = 2**64
#         for l in c:
#             max_v = max(max_v, idx[abs(l)])
#             min_v = min(min_v, idx[abs(l)])
#         for i in range(min_v, max_v):
#             cuts[i] += 1
#     print(cuts)

def TD_to_dtree(cnf, td, done = None):    
    """Constructs a (partial) dtree for a cnf from a tree decomposition.

    Build a dtree from the given tree decomposition `td` adding clauses to the dtree 
    in the order in which they are included in the tree decompositions bags. 
    Branches, when a tree decomposition bag has multiple children.
        
    Args:
        cnf (:obj:`aspmc.compile.cnf.CNF`): The cnf to construct the partial dtree for.
        td (:obj:`aspmc.graph.treedecomposition.TreeDecomposition`): A tree decomposition for a subset of the variables in the cnf.
        done (:obj:`list`, optional): A list of Booleans specifying which clauses of the cnf are already included in a dtree. 
        If given updates the list by setting the value of the added clauses to True.
        
    Returns:
        :obj:`Dtree`: The (partial) dtree corresponding to the tree decomposition.
    """
    if done == None:
        done = [ False for _ in range(len(cnf.clauses)) ]

    # at which td node to handle each clause
    clauses = {}
    # at which td node each variable occurs last
    last = {}
    idx = 0
    td_idx = list(td)
    for t in td.bag_iter():
        for a in t.vertices:
            last[a] = idx
        t.idx = idx
        idx += 1
        clauses[t] = []
    
    # set where each clause should be handled
    for i,c in enumerate(cnf.clauses):
        if not done[i]:
            if all([ abs(b) in last for b in c]):
                done[i] = True
                idx = min([ last[abs(b)] for b in c ])
                clauses[td.get_bag(td_idx[idx])].append(i)

    dtree_idx = [ None for _ in range(td.bags) ]
    for bag in td.bag_iter():
        cur_dtree = None        
        # take care of the dtrees for the children
        for child in bag.children:
            child_dtree = dtree_idx[child.idx]
            if cur_dtree == None:
                cur_dtree = child_dtree
            else:
                if child_dtree != None:
                    new_dtree = Dtree()
                    new_dtree.right = cur_dtree
                    new_dtree.left = child_dtree
                    cur_dtree = new_dtree

        # take care of the clauses that are here
        for i in clauses[bag]:
            if cur_dtree == None:
                cur_dtree = Dtree(val = i)
            else:
                new_dtree = Dtree()
                new_dtree.right = cur_dtree
                new_dtree.left = Dtree(val = i)
                cur_dtree = new_dtree

        # alternative with seemingly worse performance
        # left_dtree = binary([ dtree_idx[child.idx] for child in bag.children if dtree_idx[child.idx] is not None ])
        # right_dtree = binary([ Dtree(val = i) for i in clauses[bag] ])
        # if left_dtree is None:
        #     cur_dtree = right_dtree
        # elif right_dtree is None:
        #     cur_dtree = left_dtree
        # else:
        #     cur_dtree = Dtree()
        #     cur_dtree.left = left_dtree
        #     cur_dtree.right = right_dtree

        # remember the final dtree for this bag
        dtree_idx[bag.idx] = cur_dtree
    return dtree_idx[td.get_root().idx]

def TD_dtree(cnf, solver = "flow-cutter", timeout = "0.5"):
    """Constructs a dtree for a cnf by generating a tree decomposition and calling `TD_to_dtree` with it.

    Args:
        cnf (:obj:`aspmc.compile.cnf.CNF`): The cnf to construct the dtree for.
        solver (:obj:`string`, optional): Which solver to use to generate the tree decomposition. Defaults to "flow-cutter".
        timeout (:obj:`string`, optional): The timeout in second to give to the solver. Defaults to "0.5".
        
    Returns:
        :obj:`Dtree`: The dtree generated by the constructed tree decomposition.
    """
    td = treedecomposition.from_hypergraph(cnf.primal_hypergraph(), solver = solver, timeout=timeout)
    return TD_to_dtree(cnf, td)



class MaxTimesFloat(object):
    def __init__(self, value):
        self.value = value

    def __add__(self, other):
        return self if self.value >= other.value else other

    def __iadd__(self, other):
        self.value = max(self.value, other.value)
        return self

    def __mul__(self, other):
        return MaxTimesFloat(self.value * other.value)

    def __imul__(self, other):
        self.value *= other.value
        return self

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self)

def parse(value, atom = None):
    return MaxTimesFloat(float(value))

def from_value(value):
    return MaxTimesFloat(value)

def negate(value):
    return zero()

def to_string(value):
    return str(value.value)

def is_idempotent():
    return True

def zero():
    return MaxTimesFloat(0.0)
def one():
    return MaxTimesFloat(1.0)
dtype = object
pattern = '(1(\\.0[0]*)?|0\\.[0-9]+)'


"""
Vtree module providing methods to construct, visualize and generally work with vtrees.
"""

import aspmc.graph.treedecomposition as treedecomposition
from aspmc.graph.bintree import bintree

class Vtree(bintree):
    """A Vtree class that manages vtrees. 

    These can be used as the input to miniC2D to (exactly) specify in which order atoms are decided.
    Is a subclass of aspmc.graph.bintree.bintree and therefore has some utility functions, like writing to file, 
    counting the leaves or iterating over the nodes of the tree.
    """
    def __str__(self):
        self.index()
        res = ""
        res += f"vtree {2*self.leaf_count() - 1}\n"
        cur_max = -1
        def get_next():
            nonlocal cur_max
            cur_max += 1
            return cur_max
        for node in self:
            if node.val is not None:
                node.w_idx = get_next()
                res += f"L {node.w_idx} {node.val}\n"
            else:
                node.w_idx = get_next()
                res += f"I {node.w_idx} {node.left.w_idx} {node.right.w_idx}\n"
        return res


def from_order(order):
    """Constructs a (partial) vtree from a list of atoms in the specified order.
    Args:
        order (:obj:`list`): The list specifying which atoms should be eliminated and the order in which they should be eliminated.
        
    Returns:
        :obj:`Vtree`: The (right-linear) vtree that eliminates the atoms in `order` in the order they are in `order`.
    """
    root = Vtree()
    cur = root
    for v in order[:-1]:
        cur.left = Vtree(val = v)
        cur.right = Vtree()
        cur = cur.right
    cur.val = order[-1]
    return root

def TD_to_vtree(td):    
    """Constructs a (partial) vtree from a tree decomposition.

    Build a vtree from the given tree decomposition `td` adding variables to the dtree 
    in the order in which they are included in the tree decompositions bags. 
    Branches, when a tree decomposition bag has multiple children.
        
    Args:
        td (:obj:`aspmc.graph.treedecomposition.TreeDecomposition`): A tree decomposition.
        
    Returns:
        :obj:`Vtree`: The (partial) vtree corresponding to the tree decomposition.
    """
    # at which td node each variable occurs last
    last = {}
    # at which td node to handle which variables
    variables = { bag : [] for bag in td.bag_iter()}
    idx = 0
    td_idx = list(td.bag_iter())
    for t in td.bag_iter():
        for a in t.vertices:
            last[a] = idx
        t.idx = idx
        idx += 1
    
    for a, i in last.items():
        variables[td_idx[i]].append(a)
    

    vtree_idx = [ None for _ in range(td.bags) ]
    for bag in td.bag_iter():
        cur_vtree = None        
        # take care of the dtrees for the children
        for child in bag.children:
            child_vtree = vtree_idx[child.idx]
            if cur_vtree == None:
                cur_vtree = child_vtree
            else:
                if child_vtree != None:
                    new_vtree = Vtree()
                    new_vtree.right = cur_vtree
                    new_vtree.left = child_vtree
                    cur_vtree = new_vtree

        # take care of the clauses that are here
        for i in variables[bag]:
            if cur_vtree == None:
                cur_vtree = Vtree(val = i)
            else:
                new_vtree = Vtree()
                new_vtree.right = cur_vtree
                new_vtree.left = Vtree(val = i)
                cur_vtree = new_vtree

        vtree_idx[bag.idx] = cur_vtree
    return vtree_idx[td.get_root().idx]

def TD_vtree(cnf, solver = "htd", timeout = "0.5"):
    """Constructs a vtree for a cnf by generating a tree decomposition and calling `TD_to_vtree` with it.

    Args:
        cnf (:obj:`aspmc.compile.cnf.CNF`): The cnf to construct the dtree for.
        solver (:obj:`string`, optional): Which solver to use to generate the tree decomposition. Defaults to "htd".
        timeout (:obj:`string`, optional): The timeout in second to give to the solver. Defaults to "0.5".
        
    Returns:
        :obj:`Vtree`: The vtree generated by the constructed tree decomposition.
    """
    td = treedecomposition.from_hypergraph(cnf.primal_hypergraph(), solver = solver, timeout=timeout)
    return TD_to_vtree(td)



"""
Program module providing the nested algebraic progam class.
"""

import numpy as np

from aspmc.programs.program import Program

from aspmc.util import *

class NestedAlgebraicProgram(Program):    
    """A class for programs with weights over more than two semirings. 

    This is supposed to be an abstract class in the sense that it should used and instantiated
    by subclasses, rather than parsed itself.

    Overrides the `get_weights`, `_finalize_cnf` and `get_queries` methods appropriately.

    Args:
        clingo_control (:obj:`Control`): The clingo control object having the rules of the spanning program.
        semirings (:obj:`list[module]`): The list of semiring modules.
        weights (:obj:`list[dict]`): The list of weights over the semirings in order of the semirings. 
            Needs to map pairs `(name, phase)`, where `name` is the name of an atom 
            and `phase` is its polarity (`True` for true), to their weight.
        transforms (:obj:`list[string]`): A list of string representations of the functions that transforms a value 
            from the ith into a value from the (i-1)th semiring.
        queries (:obj:`list`): A list of names of atoms that should be queried. 
            Specify the empty list for an overall weight query.

    Attributes:
        semirings (:obj:`list[module]`): The list of semiring modules.
        weights (:obj:`list[dict]`): The list of weights over the semirings in order of the semirings. 
            Needs to map pairs `(name, phase)`, where `name` is the name of an atom 
            and `phase` is its polarity (`True` for true), to their weight.
        transforms (:obj:`list[string]`): A list of string representations of the functions that transforms a value 
            from the ith into a value from the (i-1)th semiring.
        queries (:obj:`list`): A list of names of atoms that should be queried. 
            Specify the empty list for an overall weight query.
    """
    def __init__(self, clingo_control, semirings, weights_per_semiring, transforms, queries):
        self.semirings = semirings
        self.weights_per_semiring = weights_per_semiring
        self.transforms = transforms
        self.queries = queries
        Program.__init__(self, clingo_control = clingo_control)

    def _finalize_cnf(self):
        weight_list = self.get_weights()
        varMap = { name : var for var, name in self._nameMap.items() }
        # quantifiers
        self._cnf.quantified = []
        for cur_weights in self.weights_per_semiring:
            cur_atoms = set([ int(str(varMap[name])) for (name, _) in cur_weights ])
            self._cnf.quantified.append(cur_atoms)
        # weights
        for v in range(self._cnf.nr_vars*2):
            self._cnf.weights[to_dimacs(v)] = weight_list[v]
        # rest
        self._cnf.transforms = self.transforms
        self._cnf.semirings = self.semirings
        # TODO figure out which variables we can mark as auxilliary

    def get_weights(self):
        query_cnt = max(len(self.get_queries()), 1)
        shapes = [ (query_cnt, ) + np.shape(semiring.one()) for semiring in self.semirings ]
        varMap = { name : var for var, name in self._nameMap.items() }
        # default initialize with innermost semirings one
        weight_list = [ np.empty(shapes[-1], dtype=self.semirings[-1].dtype) for _ in range(self._cnf.nr_vars*2) ]
        for i in range(self._cnf.nr_vars*2):
            weight_list[i][:] = self.semirings[-1].one()
        for i, cur_weights in enumerate(self.weights_per_semiring):
            for (name, phase) in cur_weights:
                idx = to_pos(varMap[name])
                if not phase:
                    idx = neg(idx)
                weight_list[idx] = np.empty(shapes[i], dtype=self.semirings[i].dtype)
                weight_list[idx][:] = cur_weights[(name, phase)]
                
        
        if len(self.get_queries()) == 0:
            return weight_list
        
        # set the weights for the queries
        var_to_val_type = [ len(self.semirings) - 1 for _ in range(self._cnf.nr_vars + 1) ]
        for val_type, cur_weights in enumerate(self.weights_per_semiring):
            for (name, _) in cur_weights:
                var = int(str(varMap[name]))
                var_to_val_type[var] = val_type
                
        for i, query in enumerate(self.get_queries()):
            var = int(str(varMap[query]))
            val_type = var_to_val_type[var]
            zero = self.semirings[val_type].zero()
            weight_list[neg(to_pos(varMap[query]))][i] = zero
            
        return weight_list

    def get_queries(self):
        return self.queries



def parse(f, atom = None):
    return float(f)
def from_value(value):
    return float(value)

def is_idempotent():
    return False    

def one():
    return 1.0
def zero():
    return 0.0
def negate(x):
    return 1.0 - x
def to_string(value):
    return str(value)
dtype = float
pattern = '(1(\\.0[0]*)?|0\\.[0-9]+)'



"""The aspmc.semirings module contains a collection of semiring submodules:

* grad: gradient semiring,
* maxplus: max tropical semiring,
* probabilistic: probabilistic semiring,
* two_nat: crossproduct of the natural number semiring with itself.  

They follow a simple format, thus this library can be easily extended with custom semirings.

Namely, a semiring module must provide the following functions and attributes:


    def parse(value, atom = None):
        pass

    def from_value(value):
        pass

    def negate(value):
        pass

    def to_string(value):
        pass

    def zero():
        pass

    def one():
        pass

    dtype = object
    pattern = '.*'

Here

* `parse` must take the string representation of a value and an atom and return the object corresponding to the value.
* `from_value` must take an object `value` that represents a value and return the object corresponding to the value.
    How the value object looks is up to the implementer.
* `negate` must take an object corresponding to a value and return the object corresponding to the negation of the value.
    How this negation works is up to the implementer.
* `to_string` must take an object corresponding to a value and return a string representation of it that results in the 
    same object when parsed again with `parse`.
* `zero` must return a (copy of) the object representing the zero element of the semiring.
* `one` must return a (copy of) the object representing the one element of the semiring.
* `dtype` is the type that should be used to instantiate numpy arrays of values from this semiring.
* `pattern` is a string corresponding to a regex pattern that string representations of values from this semiring
    must match.

The objects returned by `parse`, `from_value` and `negated` must implement `__add__(self, other)`, `__iadd__(self, other)`,
`__mul__(self, other)` and `__imul__(self, other)`, where `other` is an object of the same type.

Examples are in the included semiring submodules.
"""

from . import *



#!/usr/bin/env python3

"""
Maximum Expected Utility Program module specializing the TwoAlgebraicProgam class.
"""
import logging

from aspmc.parsing.clingoparser.clingoext import Control

from aspmc.programs.problogprogram import ProblogProgram
from aspmc.programs.twoalgebraicprogram import TwoAlgebraicProgram
from lark import Lark
from aspmc.parsing.lark_parser import GRAMMAR, ProblogTransformer
from aspmc.util import *

import aspmc.semirings.maxplusdecisions as first_semiring
import aspmc.semirings.grad as second_semiring # the gradient semiring and the expected utility are the same
import aspmc.semirings.probabilistic as semiring

logger = logging.getLogger("aspmc")

class MEUProblogProgram(TwoAlgebraicProgram, ProblogProgram):
    """A class for Maximum Expected Utility programs. 

    The syntax for these programs is the same as for ProbLog in MEU mode. 
    This means that the probabilistic part is specified as usual,
    utility atoms `a`, i.e. atoms whose truth adds utility value `u` to the model, must be given as `utility(a, u).` and
    the decision atoms `d`, whose assignment should be maximized over should be given as `decision(d).`.

    Subclasses `TwoAlgebraicProgram` since it is a second level problem.

    Overrides the `_prepare_grounding` method to deal with utilities and decisions.

    Args:
        program_str (:obj:`string`): A string containing a part of the program in MEU ProbLog syntax. 
        May be the empty string.
        program_files (:obj:`list`): A list of string that are paths to files which contain programs in 
        MEU ProbLog syntax that should be included. May be an empty list.
    """
    def __init__(self, program_str, program_files):
        self.semiring = semiring
        self.weights = {}
        self.utilities = {}
        self.decisions = []
        self.queries = []
        self.annotated_disjunctions = []
        for path in program_files:
            with open(path) as file_:
                program_str += file_.read()
        
        # parse the program
        my_grammar = GRAMMAR + f"""
        QUESTION_MARK : "?"
        %override weight : /{self.semiring.pattern}/ | variable | QUESTION_MARK
        """
        parser = Lark(my_grammar, start='program', parser='lalr', transformer=ProblogTransformer())
        program = parser.parse(program_str)

        # ground the program
        clingo_control = Control()
        self._ground(clingo_control, program)

        # set the weights
        first_weight_list = {}
        # weights for the maxplus semiring
        first_semiring.names = [ name for name in self.decisions ]
        for i,name in enumerate(self.decisions):
            first_weight_list[(name, True)] = first_semiring.MaxPlusFloat(0, 2**i)
            first_weight_list[(name, False)] = first_semiring.one()
        # weights for the expected utility semiring
        second_weight_list = {}
        for name in self.weights:
            second_weight_list[(name, True)] = [self.weights[name], 0.0]
            second_weight_list[(name, False)] = [1.0 - self.weights[name], 0.0]
        for (name, phase) in self.utilities:
            if (name, phase) in second_weight_list:
                if (name, phase) in first_weight_list:
                    logger.error(f"Variable {name} cannot be both a decision and probabilistic variable.")
                    exit(-1)
                second_weight_list[(name, phase)][1] = second_weight_list[(name, phase)][0]*self.utilities[(name, phase)]
            elif (name, phase) in first_weight_list:
                first_weight_list[(name, phase)].value = self.utilities[(name, phase)]
            else:
                second_weight_list[(name, phase)] = (1.0, self.utilities[(name, phase)])
        second_weight_list = { i : second_semiring.parse(f"({w[0]},{w[1]})") for (i, w) in second_weight_list.items() }
        TwoAlgebraicProgram.__init__(self, clingo_control, first_semiring, second_semiring, first_weight_list, second_weight_list, "lambda w : w.value[1]", self.queries)

    def _prepare_grounding(self, program):
        new_program = []
        for r in program:
            if isinstance(r, str):
                new_program.append(r)
            elif r.weights is not None and r.weights[0] == '?':
                atom = r.head[0]
                self.decisions.append(str(atom))
                new_program.append(f"{{{atom}}}.")
            elif r.head is not None and r.head[0].predicate == "utility":
                atom = r.head[0].inputs[0]
                utility = r.head[0].inputs[1]
                phase = True
                if atom.negated:
                    phase = False
                    atom.negated = False
                self.utilities[(str(atom), phase)] = float(utility)
            elif r.head is not None and r.head[0].predicate == "query":
                atom = r.head[0].inputs[0]
                logger.warning(f"Ignoring query for atom {atom}.")
            else:
                new_program.append(r)
        return super()._prepare_grounding(new_program)

    def _prog_string(self, program):
        result = ""
        for v in self._guess:
            if self._internal_name(v) in self.decisions:
                result += f"?::{self._external_name(v)}.\n"
            else:
                result += f"{self.weights[self._internal_name(v)]}::{self._external_name(v)}.\n"
        for (name, phase) in self.utilities:
            result += "utility("
            if not phase:
                result += "\\+"
            result += f"{name}, {self.utilities[(name, phase)]}).\n"
        for r in program:
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += ":-"
                result += ",".join([("\\+" if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        for query in self.queries:
            result += f"query({query}).\n"
        return result

    def to_pita(self):
        result = """:- use_module(library(pita)).
:- pita.
:- begin_lpad.
"""
        for v in self._guess:
            if self._internal_name(v) in self.decisions:
                result += f"? :: {self._external_name(v)}.\n"
            else:
                result += f"{self.weights[self._internal_name(v)]}::{self._external_name(v)}.\n"
        for (name, phase) in self.utilities:
            if phase:
                result += "utility("
                result += f"{name}, {self.utilities[(name, phase)]}).\n"
        for r in self._program:
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += " :- "
                result += ",".join([("\\+" if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        for query in self.queries:
            result += f"query({query}).\n"
        result += ":- end_lpad."
        return result


from typing import NamedTuple, Union, List
from dataclasses import dataclass
from clingo import Symbol # type: ignore # pylint: disable=import-error, unused-import, no-name-in-module
from aspmc.parsing.clingoparser.sign import Sign


@dataclass(eq=True, unsafe_hash=True, order=True)
class Literal:
    atom: Symbol
    sign: Sign

    def __init__(self, atom: Symbol, sign: Union[Sign, bool]):
        self.atom = atom
        if isinstance(sign, bool):
            if sign:
                sign = Sign.NoSign
            else:
                sign = Sign.Negation
        self.sign = sign

    def __repr__(self):
        return repr(self.sign) + repr(self.atom)


@dataclass(eq=True, unsafe_hash=True, order=True)
class EpistemicLiteral:
    literal: Literal
    sign: Sign = Sign.NoSign

    def __str__(self):
        return f'{self.sign}&k{{ {self.literal} }}'

    # def __init__(self, literal: Literal, sign: Sign):
    #     self.literal  = literal
    #     self.sign = sign

    # def __repr__(self):
    #     return f'{self.sign}&k{{ {self.literal} }}'

    # def __eq__(self, other):
    #     return self.literal == other.literal and self.sign == other.sign

    # def __lt__(self, other):
    #     return (self.literal, self.sign) \
    #         < (other.literal, other.sign)

class WorldView(NamedTuple):
    symbols: List[EpistemicLiteral]

    def __str__(self):
        return ' '.join(map(str, sorted(self.symbols)))



def to_dimacs(var):
    idx = var//2 + 1
    if var % 2 == 1:
        idx *= -1
    return idx

def to_pos(var):
    if var > 0:
        return 2*(var - 1)
    else:
        return 2*(-var - 1) + 1

def neg(var, dimacs = False):
    if dimacs:
        return -var
    else:
        return var ^ 1


#!/usr/bin/env python3

"""
Program module providing the smprogam class.
"""
import logging
import numpy as np

from aspmc.parsing.clingoparser.clingoext import Control

from aspmc.programs.problogprogram import ProblogProgram
from aspmc.programs.twoalgebraicprogram import TwoAlgebraicProgram
from lark import Lark
from aspmc.parsing.lark_parser import GRAMMAR, ProblogTransformer, ProbabilisticRule, Atom
from aspmc.util import *

import aspmc.semirings.probabilistic as first_semiring
import aspmc.semirings.two_nat as second_semiring

logger = logging.getLogger("aspmc")

class SMProblogProgram(TwoAlgebraicProgram, ProblogProgram):    
    """A class for SMProbLog programs. 

    Subclasses `TwoAlgebraicProgram` since it is a second level problem.

    Overrides the `_prepare_grounding` method to deal with negative atoms in the head in the same way as ProbLog.

    Args:
        program_str (:obj:`string`): A string containing a part of the program in ProbLog syntax. 
        May be the empty string.
        program_files (:obj:`list`): A list of string that are paths to files which contain programs in 
        ProbLog syntax that should be included. May be an empty list.
    """
    def __init__(self, program_str, program_files):
        self.semiring = first_semiring
        self.weights = {}
        self.queries = []
        self.annotated_disjunctions = []
        for path in program_files:
            with open(path) as file_:
                program_str += file_.read()

        # parse the program
        my_grammar = GRAMMAR + f"%override weight : /{self.semiring.pattern}/ | variable\n"
        parser = Lark(my_grammar, start='program', parser='lalr', transformer=ProblogTransformer())
        program = parser.parse(program_str)

        # ground the program
        clingo_control = Control()
        self._ground(clingo_control, program)

        # initialize weights
        weight_list = {}
        for name in self.weights:
            weight_list[(name, True)] = self.weights[name]
            weight_list[(name, False)] = 1 - self.weights[name]

        TwoAlgebraicProgram.__init__(self, clingo_control, first_semiring, second_semiring, weight_list, {}, "lambda w : w[0]/w[1]", self.queries)

    def _prepare_grounding(self, program):
        # take care of the transformation for negated head atoms
        # first gather all the head atoms grouped by whether they are negated and their predicate
        program = list(program)
        predicate_to_pos_heads = {}
        predicate_to_neg_heads = {}
        for r in program:
            if isinstance(r, str):
                program.append(r)
            elif r.head is not None:
                for a in r.head:
                    if a.negated:
                        if not a.predicate in predicate_to_neg_heads:
                            predicate_to_neg_heads[a.predicate] = []
                        predicate_to_neg_heads[a.predicate].append(a)
                    else:
                        if not a.predicate in predicate_to_pos_heads:
                            predicate_to_pos_heads[a.predicate] = []
                        predicate_to_pos_heads[a.predicate].append(a)
        
        
        # next, for every predicate pred that occurs negated introduce two new predicates pos_pred, neg_pred
        # derive pred only of pos_pred but not neg_pred holds
        for pred in predicate_to_neg_heads:
            arities = set()
            for atom in predicate_to_neg_heads[pred]:
                atom.predicate = f"neg_{pred}"
                atom.negated = False
                arities.add(len(atom.inputs))
            for atom in predicate_to_pos_heads[pred]:
                if len(atom.inputs) in arities: # otherwise we can trust that we really derive the positive version anyway
                    atom.predicate = f"pos_{pred}"
            for l in arities:
                head = [ Atom(pred, [ f"X{i}" for i in range(l) ], negated = False) ]
                body = [ 
                    Atom(f"pos_{pred}", [ f"X{i}" for i in range(l) ], negated = False),  
                    Atom(f"neg_{pred}", [ f"X{i}" for i in range(l) ], negated = True)
                    ]
                rule = ProbabilisticRule(head, body, weights = None)
                program.append(rule)
        return super()._prepare_grounding(program)

    def _prog_string(self, program):
        result = ""
        for v in self._guess:
            result += f"{self.weights[self._internal_name(v)]}::{self._external_name(v)}.\n"
        for r in program:
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += ":-"
                result += ",".join([("\\+ " if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        for query in self.queries:
            result += f"query({query}).\n"
        return result

    def get_weights(self):
        varMap = { name : var for var, name in self._nameMap.items() }
        weight_list = super().get_weights()
        for i, query in enumerate(self.queries):
            if (query, True) in self.first_weights:
                weight_list[neg(to_pos(varMap[query]))][i] = self.first_semiring.zero()
            else:
                weight_list[neg(to_pos(varMap[query]))][i] = np.array([0.0, 1.0])
        return weight_list


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import generator_stop

import sys

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu, leftrec, nomemo
from tatsu.parsing import leftrec, nomemo  # noqa
from tatsu.util import re, generic_main  # noqa


KEYWORDS = {}  # type: ignore


class WeightedFormulaBuffer(Buffer):
    def __init__(
        self,
        text,
        whitespace=None,
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        namechars='',
        **kwargs
    ):
        super().__init__(
            text,
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            namechars=namechars,
            **kwargs
        )


class WeightedFormulaParser(Parser):
    def __init__(
        self,
        whitespace=None,
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        left_recursion=True,
        parseinfo=True,
        keywords=None,
        namechars='',
        tokenizercls=WeightedFormulaBuffer,
        **kwargs
    ):
        if keywords is None:
            keywords = KEYWORDS
        super().__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            parseinfo=parseinfo,
            keywords=keywords,
            namechars=namechars,
            tokenizercls=tokenizercls,
            **kwargs
        )

    @tatsumasu()
    @nomemo
    def _start_(self):  # noqa
        self._expression_()
        self._check_eof()

    @tatsumasu()
    @leftrec
    def _expression_(self):  # noqa
        with self._choice():
            with self._option():
                self._expression_()
                self._token('+')
                self._term_()
            with self._option():
                self._term_()
            self._error('expecting one of: expression factor term')

    @tatsumasu()
    @leftrec
    def _term_(self):  # noqa
        with self._choice():
            with self._option():
                self._term_()
                self._token('*')
                self._factor_()
            with self._option():
                self._factor_()
            self._error('expecting one of: ( factor number term var')

    @tatsumasu()
    def _factor_(self):  # noqa
        with self._choice():
            with self._option():
                self._token('(')
                self._expression_()
                self._token(')')
            with self._option():
                self._number_()
            with self._option():
                self._var_()
            self._error('expecting one of: ( /#\\([^\\(+*,\\)]*\\)/ /[a-z][^\\(+*,\\)]*/ not  number var')

    @tatsumasu()
    def _var_(self):  # noqa
        with self._optional():
            self._token('not ')
        self._pattern('[a-z][^\\(+*,\\)]*')
        with self._optional():
            self._pattern('\\([^\\(+*\\)]*\\)')

    @tatsumasu()
    def _number_(self):  # noqa
        self._pattern('#\\([^\\(+*,\\)]*\\)')


class WeightedFormulaSemantics(object):
    def __init__(self, app):
        self._app = app

    def var2idx(self, var):
        return self._app.var2idx(var)

    def new_var(self):
        return self._app.new_var("")

    def clause_writer(self, p, c1 = 0, c2 = 0, connective = 0):
        return self._app.clause_writer(p, c1 = c1, c2 = c2, connective = connective)

    def start(self, ast):  # noqa
        self._app._clauses.append([ast])
        return ast

    def expression(self, ast):  # noqa
        if type(ast) == tuple and "+" in ast:
            n_var = self.new_var()
            self._app.clause_writer(n_var, c1 = ast[0], c2 = ast[2], connective = 5)
            return n_var
        return ast

    def term(self, ast):  # noqa
        if type(ast) == tuple and "*" in ast:
            n_var = self.new_var()
            self._app.clause_writer(n_var, c1 = ast[0], c2 = ast[2], connective = 4)
            return n_var
        return ast

    def factor(self, ast):  # noqa
        if type(ast) == tuple and ast[0] == "(":
            ast = ast[1]
        return ast

    def var(self, ast):  # noqa
        if type(ast) == tuple:
            act = ""
            for i in range(len(ast)):
                act += ast[i]
            ast = act
        if len(ast) > 3 and ast[:3] == "not":
            idx = -self.var2idx(ast[:3])
            # if the variable was optimized away, we know it is false
            if idx == 0:
                return self.number("#(1)")
        else:
            idx = self.var2idx(ast)
            # if the variable was optimized away, we know it is false
            if idx == 0:
                return self.number("#(0)")
        n_var = self.new_var()
        self._app._clauses.append([-n_var, idx])
        return n_var

    def number(self, ast):  # noqa
        n_var = self.new_var()
        self._app._weights[n_var] = ast[2:-1]
        return n_var



"""Compilation module of aspmc.

This module provides access to an extended cnf class, which can be used to compile and 
evaluate algebraic model counting problems over cnfs using the knowledge compiler in aspmc.config.

CNF class:
    The module aspmc.compile.cnf provides an extended cnf class, which also stores:

    * weights of literals
    * the semiring(s) over which the weights are
    * information about which variables are "quantified" over which semiring
    * how to convert the value of one semiring value into another one

    Most prominently it has a compile method, which uses the information of the extended cnf
    to evaluate the algebraic model counting problem described in it.  

XTrees:
    The modules aspmc.compile.dtree and aspmc.compile.vtree contain a class Dtree and Vtree respectively.
    They can be given to c2d and miniC2D respectively to tell them in which order variables should be decided.
    Both trees can be generated from treedecompositions to obtain trees that allow for efficient evaluation.

Evaluation:
    The modules aspmc.compile.circuit, aspmc.compile.constrained_ddnnf, aspmc.compile.constrained_sdd contain
    different circuit classes, each of which have static methods for parsing a tractalble circuit and evaluating an 
    algebraic model counting instance over it in polynomial time in the size of the circuit. The different submodules handle different circuits:

    * aspmc.compile.circuit handles first level problems, where only one semiring is used.
    * aspmc.compile.constrained_ddnnf handles second level problems, where two semirings are used, over (smooth) ddnnfs.
    * aspmc.compile.constrained_sdd handles second level problems, where two semirings are used, over sdds.

Separators:
    The module aspmc.compile.separator contains a clingo extension that computes an optimal model of an ASP program with #optimize statements.
"""

from . import *



names = []

class MinMaxPlusFloat(object):
    def __init__(self, value, decisions):
        self.value = value
        self.decisions = decisions

    def __add__(self, other):
        first = self if self.value[0] >= other.value[0] else other
        second = self if self.value[1] >= other.value[1] else other
        return MinMaxPlusFloat((first.value[0], second.value[1]), (first.decisions[0], second.decisions[1]))

    def __iadd__(self, other):
        if self.value[0] > other.value[0]:
            self.value[0] = other.value[0]
            self.decisions[0] = other.decisions[0]
        if self.value[1] > other.value[1]:
            self.value[1] = other.value[1]
            self.decisions[1] = other.decisions[1]
        return self

    def __mul__(self, other):
        return MinMaxPlusFloat( (self.value[0] + other.value[0], self.value[1] + other.value[1]),
                                (self.decisions[0] | other.decisions[0], self.decisions[1] | other.decisions[1]))

    def __imul__(self, other):
        self.value[0] += other.value[0]
        self.value[1] += other.value[1]
        self.decisions[0] |= other.decisions[0]
        self.decisions[1] |= other.decisions[1]
        return self

    def __str__(self):
        min_decisions = [ names[i] for i in range(len(names)) if self.decisions[0] & 2**i ]
        min_decisions = ", ".join(min_decisions)
        max_decisions = [ names[i] for i in range(len(names)) if self.decisions[1] & 2**i ]
        max_decisions = ", ".join(max_decisions)
        ret = f"Minimum of {self.value[0]} with true atoms: {min_decisions}\n"
        ret += f"Maximum of {self.value[1]} with true atoms: {max_decisions}"
        return ret

    def __repr__(self):
        return str(self)

def parse(value, atom = None):
    value = value[1:-1]
    value = value.split(',')
    return MinMaxPlusFloat((float(value[0]), float(value[1])), (int(value[2]), int(value[3])))

def from_value(value):
    return MinMaxPlusFloat(value, (0, 0))

def negate(value):
    return one()

def to_string(value):
    return f"({value.value},{value.decisions})"

def is_idempotent():
    return True

def zero():
    return MinMaxPlusFloat((float("-inf"), float("-inf")), (0, 0))
def one():
    return MinMaxPlusFloat((0, 0), (0, 0))

dtype = object
pattern = '\(\([+-]?([0-9]*[.])?[0-9]+,[+-]?([0-9]*[.])?[0-9]+,\),\([0-9]+,[0-9]+\)\)'


class TwoGradient(object):
    def __init__(self, value):
        self.value = value

    def __add__(self, other):
        return TwoGradient((self.value[0] + other.value[0], self.value[1] + other.value[1], self.value[2] + other.value[2]))

    def __iadd__(self, other):
        self.value[0] += other.value[0]
        self.value[1] += other.value[1]
        self.value[2] += other.value[2]
        return self

    def __mul__(self, other):
        return TwoGradient((self.value[0] * other.value[0], 
                            self.value[0] * other.value[1] + self.value[1] * other.value[0], 
                            self.value[0] * other.value[2] + self.value[2] * other.value[0]))

    def __imul__(self, other):
        self.value[1] *= other.value[0]
        self.value[1] += self.value[0] * other.value[1]
        self.value[2] *= other.value[0]
        self.value[2] += self.value[0] * other.value[2]
        self.value[0] *= other.value[0]
        return self

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self.value)

def parse(value, atom = None):
    value = value[1:-1]
    value = value.split(',')
    return TwoGradient((float(value[0]), float(value[1]), float(value[2])))

def from_value(value):
    return TwoGradient(value)

def negate(value):
    return TwoGradient((1.0 - value.value[0], -value.value[1], -value.value[2]))

def to_string(value):
    return f"({value.value[0]},{value.value[1]},{value.value[2]})"

def is_idempotent():
    return False

def zero():
    return TwoGradient((0.0, 0.0, 0.0))
def one():
    return TwoGradient((1.0, 0.0, 0.0))

dtype = object
pattern = '\([+-]?([0-9]*[.])?[0-9]+,[+-]?([0-9]*[.])?[0-9],[+-]?([0-9]*[.])?[0-9]+\)'



"""
Program module providing the algebraic progam class.
"""

import re
import numpy as np
import logging

from aspmc.programs.program import Program, UnsupportedException

from aspmc.parsing.clingoparser.clingoext import ClingoRule, Control

from lark import Lark
from aspmc.parsing.lark_parser import GRAMMAR, ProblogTransformer
import aspmc.programs.grounder as grounder

import aspmc.semirings.probabilistic as semiring

from aspmc.util import *
from aspmc.programs.naming import *

logger = logging.getLogger("aspmc")

class ProblogProgram(Program):
    """A class for programs with weights over semirings. 

    Should be specified in ProbLog syntax, but allows negations and negative cycles.

    For programs over semirings that are not the probabilistic one, see aspmc.programs.algebraicprogram.

    Grounding of these programs (and subclasses thereof) should follow the following strategy:

    * `_prepare_grounding(self, program)` should take the output of the parser 
        (i.e. a list of rules and special objects) and process all the rules and special objects
        transforming them either into other rules or into strings that can be given to the grounder.
    * the output of `_prepare_grounding(self, program)` is transformed to one program string via

            '\\n'.join([ str(r) for r in program ])
        
        This string will be given to the grounder, which produces a clingo control object.
    * `_process_grounding(self, clingo_control)` should take this clingo control object and process the
        grounding in an appropriate way (and draw some information from it optionally about weights, special objects).
        The resulting processed clingo_control object must only know about the 
        rules that should be seen in the base program class.

    Thus, subclasses can override `_prepare_grounding` and `_process_grounding` (and optionally call the superclass methods) 
    to handle their extras. See aspmc.programs.meuprogram or aspmc.programs.smprogram for examples.

    Args:
        program_str (:obj:`string`): A string containing a part of the program in ProbLog syntax. 
        May be the empty string.
        program_files (:obj:`list`): A list of string that are paths to files which contain programs in 
        ProbLog syntax that should be included. May be an empty list.

    Attributes:
        semiring (:obj:`module`): The semiring module to be used. 
        weights (:obj:`dict`): The dictionary from atom names to their weight.
        queries (:obj:`list`): The list of atoms to be queries in their string representation.
    """
    def __init__(self, program_str, program_files):
        self.semiring = semiring
        self.weights = {}
        self.queries = []
        self.annotated_disjunctions = []
        for path in program_files:
            with open(path) as file_:
                program_str += file_.read()
        # parse
        my_grammar = GRAMMAR + f"%override weight : /{self.semiring.pattern}/ | variable\n"
        parser = Lark(my_grammar, start='program', parser='lalr', transformer=ProblogTransformer())
        program = parser.parse(program_str)

        # ground
        clingo_control = Control()
        self._ground(clingo_control, program)

        # initialize the superclass
        Program.__init__(self, clingo_control = clingo_control)

    def _ground(self, clingo_control, program):
        # do the grounding in three steps:
        # 1. transform the parsed rules into asp rules
        program = self._prepare_grounding(program)
        # 2. give the asp rules to the grounder
        clingo_str = '\n'.join([ str(r) for r in program ])
        grounder.ground(clingo_control, program_str = clingo_str, program_files = [])
        # 3. take care of possible extras
        self._process_grounding(clingo_control)

    def _prepare_grounding(self, program):
        guess_idx = 0
        new_program = []
        for r in program:
            if isinstance(r, str):
                new_program.append(r)
            elif r.head is not None and r.head[0].predicate == "query":
                atom = r.head[0].inputs[0]
                new_program.append(f"query_atom({atom}):-{atom}.")
            elif r.weights is not None:
                self.annotated_disjunctions.append({})
                variables = set()
                for atom in r.body:
                    variables.update(atom.get_variables())
                if len(variables) == 0:
                    variables.add("none")
                variables = f"set({','.join([ str(x) for x in variables ])})"
                # take care of the first atom
                guess_atom = f"{INTERNAL_ALGEBRAIC}({guess_idx},{0},{0},{variables},{r.head[0]},"
                if re.match(r"[A-Z][a-zA-Z0-9]*", str(r.weights[0])):
                    guess_atom += str(r.weights[0]) + ")"
                else:
                    guess_atom += f"\"{r.weights[0]}\")"
                guess_rule = f"{{{guess_atom}}}:-{','.join([str(x) for x in r.body])}."
                use_rule = f"{r.head[0]}:-{','.join([str(x) for x in r.body]+[guess_atom])}."
                new_program += [ guess_rule, use_rule ]
                self.annotated_disjunctions[guess_idx]
                prev = guess_atom
                for i in range(1, len(r.weights)): # take care of the rest iteratively
                    guess_atom = f"{INTERNAL_ALGEBRAIC}({guess_idx},{i},{0},{variables},{r.head[i]},"
                    prev_atom = f"{INTERNAL_ALGEBRAIC}({guess_idx},{i},{1},{variables},{r.head[i]},"
                    if re.match(r"[A-Z][a-zA-Z0-9]*", str(r.weights[i])):
                        guess_atom += str(r.weights[i]) + ")"
                        prev_atom += str(r.weights[i]) + ")"
                    else:
                        guess_atom += f"\"{r.weights[i]}\")"
                        prev_atom += f"\"{r.weights[i]}\")"
                    guess_rule = f"{{{guess_atom}}}:-{','.join([str(x) for x in r.body])}."
                    use_rule = f"{r.head[i]}:-{','.join([str(x) for x in r.body]+['not '+prev, guess_atom])}."
                    prev_rule_1 = f"{prev_atom}:-{','.join([str(x) for x in r.body] + [prev])}."
                    prev_rule_2 = f"{prev_atom}:-{','.join([str(x) for x in r.body] + [guess_atom])}."
                    new_program += [ guess_rule, use_rule, prev_rule_1, prev_rule_2 ]
                    prev = prev_atom
                guess_idx += 1
            else:
                new_program.append(r)
        return new_program

    def _process_grounding(self, clingo_control):
        new_objects = []
        symbol_map = {}
        conditioned = {}
        # remember for each atom which rules can derive it
        # if there is only one such rule and it is a guess, then we want to transfer the guess to the atom
        per_head = {}
        # remember the true atoms so we can check verify that it makes sense that there are no conditional rules that derive an atom
        trues = set()
        for sym in clingo_control.symbolic_atoms:
            symbol_map[sym.literal] = str(sym.symbol)
        for o in clingo_control.ground_program.objects:
            if isinstance(o, ClingoRule):
                head_name = "" if len(o.head) == 0 else symbol_map[abs(o.head[0])]
                # check if the rule corresponds to a query
                if head_name.startswith("query_atom"):
                    if len(o.body) > 0:
                        self.queries.append(symbol_map[abs(o.body[0])])
                    else:
                        atom = head_name[11:-1]
                        logger.warning(f"Query for atom {atom} was proven true during grounding.")
                        logger.warning(f"Including it has a negative impact on performance.")
                        self.queries.append(atom)
                elif head_name.startswith(INTERNAL_ALGEBRAIC):
                    # find out the index of the annotated disjunction
                    cur_idx = len(INTERNAL_ALGEBRAIC) + 1
                    next_idx = head_name.find(",", cur_idx)
                    guess_idx = int(head_name[cur_idx:next_idx])
                    cur_idx = next_idx + 1
                    next_idx = head_name.find(",", cur_idx)
                    iter_idx = int(head_name[cur_idx:next_idx])
                    cur_idx = next_idx + 1
                    next_idx = head_name.find(",", cur_idx)
                    type_idx = int(head_name[cur_idx:next_idx])
                    # find out which terms were used for grounding
                    cur_idx = next_idx + 5 # == len(",set(")
                    end_idx = head_name.find(')', cur_idx)
                    # now cur_idx is the location of the closing bracket
                    variables = tuple(head_name[cur_idx: end_idx].split(","))
                    if variables not in self.annotated_disjunctions[guess_idx]:
                        self.annotated_disjunctions[guess_idx][variables] = []
                    while len(self.annotated_disjunctions[guess_idx][variables]) <= iter_idx:
                        self.annotated_disjunctions[guess_idx][variables].append([None, []])
                    if type_idx == 0: # guess rule
                        self.annotated_disjunctions[guess_idx][variables][iter_idx][type_idx] = o
                    else: # prev rule
                        self.annotated_disjunctions[guess_idx][variables][iter_idx][type_idx].append(o)
                # check if the rule has algebraic_atom in the body
                # these rules are handled later
                elif any([ a > 0 and symbol_map[a].startswith(INTERNAL_ALGEBRAIC) for a in o.body ]):
                    for a in o.body:
                        if a > 0 and symbol_map[a].startswith(INTERNAL_ALGEBRAIC):
                            conditioned[a] = o
                else:
                    new_objects.append(o)
                
                if len(o.head) == 1:
                    if o.head[0] not in per_head:
                        per_head[o.head[0]] = []
                    per_head[o.head[0]].append(o)

                if len(o.head) == 1 and len(o.body) == 0 and not o.choice:
                    trues.add(o.head[0])

        for idx in range(len(self.annotated_disjunctions)):
            for variables in self.annotated_disjunctions[idx]:
                rules = self.annotated_disjunctions[idx][variables]
                if len(rules) == 0:
                    logger.error("There must be at least one probabilistic atom in an annotated disjuntion.")
                    exit(-1)
                if len(rules) > 1: # handle proper annotated disjunctions
                    rest = self.semiring.one()
                    for rule in rules:
                        for prev_rule in rule[1]:
                            # for prev we do not need the other body atoms
                            prev_rule.body = [ a for a in prev_rule.body if a > 0 and symbol_map[a].startswith(INTERNAL_ALGEBRAIC)] 
                            new_objects.append(prev_rule)
                        rule[0].body = [] # make the guess unconditional
                        new_objects.append(rule[0])
                        if not rule[0].head[0] in conditioned:
                            found = False
                            for true_atom in trues:
                                true_atom_name = symbol_map[true_atom]
                                if f",{true_atom_name}," in symbol_map[rule[0].head[0]]:
                                    found = True
                                    break
                            assert(found)
                            # if the atom was proven to be true during grounding there will be no conditioned rule
                        else:
                            new_objects.append(conditioned[rule[0].head[0]])
                        # find out the weight
                        head_name = symbol_map[rule[0].head[0]]
                        start = len(head_name) - 3
                        while head_name[start] != "\"":
                            start -= 1
                        start += 1
                        weight = self.semiring.parse(head_name[start:-2])
                        tmp = rest - weight
                        if rest < 0.000001:
                            if weight > 0.000001:
                                logger.error("Probabilities that do not sum up to one!")
                                exit(-1)
                        else:
                            weight /= rest
                        self.weights[head_name] = max(min(weight, 1.0), 0.0) 
                        rest = tmp
                else: # handle single atom guesses
                    rule = rules[0]
                    if not rule[0].head[0] in conditioned:
                        found = False
                        for true_atom in trues:
                            true_atom_name = symbol_map[true_atom]
                            if f",{true_atom_name}," in symbol_map[rule[0].head[0]]:
                                found = True
                                break
                        assert(found)
                        # if the atom was proven to be true during grounding there will be no conditioned rule
                        actual_name = symbol_map[rule[0].head[0]]
                        rule[0].body = [] # make the guess unconditional
                        new_objects.append(rule[0])
                    else:
                        conditioned_rule = conditioned[rule[0].head[0]]
                        # if the conditioned_rule is the only rule deriving this atom we transfer the guess to it
                        if len(conditioned_rule.head) == 1 and len(per_head[conditioned_rule.head[0]]) == 1 and len(conditioned_rule.body) == 1:
                            actual_name = symbol_map[conditioned_rule.head[0]]
                            conditioned_rule.body = []
                            conditioned_rule.choice = True
                            new_objects.append(conditioned_rule)
                        else:
                            actual_name = symbol_map[rule[0].head[0]]
                            rule[0].body = [] # make the guess unconditional
                            new_objects.append(rule[0])
                            new_objects.append(conditioned_rule)
                        # find out the weight
                    head_name = symbol_map[rule[0].head[0]]
                    start = len(head_name) - 3
                    while head_name[start] != "\"":
                        start -= 1
                    start += 1
                    weight = self.semiring.parse(head_name[start:-2])
                    self.weights[actual_name] = weight

        clingo_control.ground_program.objects = new_objects

    def _prog_string(self, program):
        result = ""
        for v in self._guess:
            result += f"{self.weights[self._internal_name(v)]}::{self._external_name(v)}.\n"
        for r in program:
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += ":-"
                result += ",".join([("\\+ " if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        for query in self.queries:
            result += f"query({query}).\n"
        return result

    def to_lpmln(self):
        import math
        result = ""
        for v in self._guess:
            result += f"{{{self._external_name(v)}}}.\n"
        for r in self._program:
            result += ";".join([self._external_name(v) for v in r.head])
            if len(r.body) > 0:
                result += ":-"
                result += ",".join([("not " if v < 0 else "") + self._external_name(abs(v)) for v in r.body])
            result += ".\n"
        for query in self.queries:
            result += f"query({query}) :- {query}.\n"
        for v in self._guess:
            if self.weights[self._internal_name(v)] <= 0.0:
                result += f":- {self._external_name(v)}.\n"
            elif self.weights[self._internal_name(v)] >= 1.0:
                result += f":- not {self._external_name(v)}.\n"
            else:
                if abs(math.log(self.weights[self._internal_name(v)])) < 0.1**10:
                    result += f":- {self._external_name(v)}.\n"
                elif abs(math.log(1.0 - self.weights[self._internal_name(v)])) < 0.1**10:
                    result += f":- not {self._external_name(v)}.\n"
                result += f"{math.log(self.weights[self._internal_name(v)])}:- not {self._external_name(v)}.\n"
                result += f"{math.log(1.0 - self.weights[self._internal_name(v)])}:- {self._external_name(v)}.\n"
        return result

    def _finalize_cnf(self):
        weight_list = self.get_weights()
        for v in range(self._cnf.nr_vars*2):
            self._cnf.weights[to_dimacs(v)] = weight_list[v]
        self._cnf.semirings = [ self.semiring ]
        self._cnf.quantified = [ list(range(1, self._cnf.nr_vars + 1)) ]
        query_cnt = max(len(self.queries), 1)
        for v in self._deriv:
            if all(self._cnf.weights[v][i] == self.semiring.one() and self._cnf.weights[-v][i] == self.semiring.one() for i in range(query_cnt)):
                self._cnf.auxilliary.add(v)

    def get_weights(self):
        query_cnt = max(len(self.queries), 1)
        varMap = { name : var for var, name in self._nameMap.items() }
        weight_list = [ np.full(query_cnt, self.semiring.one(), dtype=self.semiring.dtype) for _ in range(self._cnf.nr_vars*2) ]
        for name in self.weights:
            weight_list[to_pos(varMap[name])] = np.full(query_cnt, self.weights[name], dtype=self.semiring.dtype)
            weight_list[neg(to_pos(varMap[name]))] = np.full(query_cnt, self.semiring.negate(self.weights[name]), dtype=self.semiring.dtype)
        for i, query in enumerate(self.queries):
            weight_list[neg(to_pos(varMap[query]))][i] = self.semiring.zero()
        return weight_list

    def get_queries(self):
        return self.queries


class MinMaxPlusFloat(object):
    def __init__(self, value):
        self.value = value

    def __add__(self, other):
        return MinMaxPlusFloat((min(self.value[0], other.value[0]), max(self.value[1], other.value[1])))

    def __iadd__(self, other):
        self.value[0] = max(self.value[0], other.value[0])
        self.value[1] = max(self.value[1], other.value[1])
        return self

    def __mul__(self, other):
        return MinMaxPlusFloat((self.value[0] + other.value[0], self.value[1] + other.value[1]))

    def __imul__(self, other):
        self.value[0] += other.value[0]
        self.value[1] += other.value[1]
        return self

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self)

def parse(value, atom = None):
    value = value[1:-1]
    value = value.split(',')
    return MinMaxPlusFloat((float(value[0]), float(value[1])))

def from_value(value):
    return MinMaxPlusFloat(value)

def negate(value):
    return zero()

def to_string(value):
    return str(value.value)
    
def is_idempotent():
    return True

def zero():
    return MinMaxPlusFloat((float("inf"), float("-inf")))

def one():
    return MinMaxPlusFloat((0, 0))
dtype = object
pattern = '\([+-]?([0-9]*[.])?[0-9]+,[+-]?([0-9]*[.])?[0-9]+\)'
